from os import getcwd
import os
from pathlib import Path

import numpy as np
import pytest
from click.testing import CliRunner

from hangar import Repository
from hangar.cli import cli
from hangar.external import PluginManager
from conftest import fixed_shape_backend_params


# -------------------------------- test data ----------------------------------


help_res = 'Usage: main [OPTIONS] COMMAND [ARGS]...\n'\
           '\n'\
           'Options:\n'\
           '  --version  display current Hangar Version\n'\
           '  --help     Show this message and exit.\n'\
           '\n'\
           'Commands:\n'\
           '  branch       operate on and list branch pointers.\n'\
           '  checkout     Checkout writer head branch at BRANCHNAME.\n'\
           '  clone        Initialize a repository at the current path and fetch updated...\n'\
           '  column       Operations for working with columns in the writer checkout.\n'\
           '  commit       Commits outstanding changes.\n'\
           '  diff         Display diff of DEV commit/branch to MASTER commit/branch.\n'\
           '  export       Export COLUMN sample data as it existed a STARTPOINT to some...\n'\
           '  fetch        Retrieve the commit history from REMOTE for BRANCH.\n'\
           '  fetch-data   Get data from REMOTE referenced by STARTPOINT (short-commit or...\n'\
           '  import       Import file or directory of files at PATH to COLUMN in the...\n'\
           '  init         Initialize an empty repository at the current path\n'\
           '  log          Display commit graph starting at STARTPOINT (short-digest or...\n'\
           '  push         Upload local BRANCH commit history / data to REMOTE server.\n'\
           '  remote       Operations for working with remote server references\n'\
           '  server       Start a hangar server, initializing one if does not exist.\n'\
           '  status       Display changes made in the staging area compared to it\'s base...\n'\
           '  summary      Display content summary at STARTPOINT (short-digest or branch).\n'\
           '  view         Use a plugin to view the data of some SAMPLE in COLUMN at...\n'\
           '  writer-lock  Determine if the writer lock is held for a repository.\n'


# ------------------------------- begin tests ---------------------------------


def test_help_option():
    runner = CliRunner()
    with runner.isolated_filesystem():
        res = runner.invoke(cli.main, ['--help'], terminal_width=80)
        assert res.exit_code == 0
        assert res.stdout == help_res


def test_help_no_args_option():
    runner = CliRunner()
    with runner.isolated_filesystem():
        res = runner.invoke(cli.main, terminal_width=80)
        assert res.exit_code == 0
        assert res.stdout == help_res


def test_version_long_option():
    import hangar
    runner = CliRunner()
    with runner.isolated_filesystem():
        res = runner.invoke(cli.main, ['--version'])
        assert res.exit_code == 0
        assert res.stdout == f'main, version {hangar.__version__}\n'


def test_init_repo(managed_tmpdir):
    runner = CliRunner()
    with runner.isolated_filesystem():
        P = getcwd()
        try:
            repo = Repository(P, exists=False)
            res = runner.invoke(cli.init, ['--name', 'test', '--email', 'test@foo.com'], obj=repo)
            assert res.exit_code == 0
            assert repo._Repository__verify_repo_initialized() is None
        finally:
            repo._env._close_environments()


def test_writer_lock_is_held_check(repo_20_filled_samples2):
    runner = CliRunner()
    res = runner.invoke(cli.writer_lock_held, obj=repo_20_filled_samples2)
    assert res.exit_code == 0
    assert res.stdout == 'Writer lock is available.\n'
    co = repo_20_filled_samples2.checkout(write=True)
    res = runner.invoke(cli.writer_lock_held, obj=repo_20_filled_samples2)
    assert res.exit_code == 0
    assert res.stdout == 'Writer lock is held.\n'
    co.close()


def test_writer_lock_force_release(repo_20_filled_samples2):
    runner = CliRunner()
    res = runner.invoke(cli.writer_lock_held, ['--force-release'], obj=repo_20_filled_samples2)
    assert res.exit_code == 0
    assert res.stdout == 'Success force release of writer lock.\n'
    co = repo_20_filled_samples2.checkout(write=True)
    res = runner.invoke(cli.writer_lock_held, ['--force-release'], obj=repo_20_filled_samples2)
    assert res.exit_code == 0
    assert res.stdout == 'Success force release of writer lock.\n'
    assert repo_20_filled_samples2.writer_lock_held is False
    nco = repo_20_filled_samples2.checkout(write=True)
    with pytest.raises(PermissionError):
        print(co.columns)
    nco.close()


def test_checkout_writer_branch_works(repo_20_filled_samples2):
    from hangar.records.heads import get_staging_branch_head
    repo_20_filled_samples2.create_branch('dev')
    runner = CliRunner()
    res = runner.invoke(cli.checkout, ['dev'], obj=repo_20_filled_samples2)
    assert res.exit_code == 0
    assert res.stdout == 'Writer checkout head set to branch: dev\n'
    recorded_branch = get_staging_branch_head(repo_20_filled_samples2._env.branchenv)
    assert recorded_branch == 'dev'
    assert repo_20_filled_samples2.writer_lock_held is False


def test_checkout_writer_branch_nonexistant_branch_errors(repo_20_filled_samples2):
    from hangar.records.heads import get_staging_branch_head
    runner = CliRunner()
    res = runner.invoke(cli.checkout, ['doesnotexist'], obj=repo_20_filled_samples2)
    assert res.exit_code == 1
    assert res.stdout == 'Error: branch with name: doesnotexist does not exist. cannot get head.\n'
    recorded_branch = get_staging_branch_head(repo_20_filled_samples2._env.branchenv)
    assert recorded_branch == 'master'
    assert repo_20_filled_samples2.writer_lock_held is False


def test_checkout_writer_branch_lock_held_errors(repo_20_filled_samples2):
    from hangar.records.heads import get_staging_branch_head
    repo_20_filled_samples2.create_branch('testbranch')
    co = repo_20_filled_samples2.checkout(write=True, branch='master')
    try:
        runner = CliRunner()
        res = runner.invoke(cli.checkout, ['testbranch'], obj=repo_20_filled_samples2)
        assert res.exit_code == 1
        msg = res.stdout
        assert msg.startswith('Error: Cannot acquire the writer lock.') is True
        recorded_branch = get_staging_branch_head(repo_20_filled_samples2._env.branchenv)
        assert recorded_branch == 'master'
        assert repo_20_filled_samples2.writer_lock_held is True
        assert co.branch_name == 'master'
    finally:
        co.close()
    assert repo_20_filled_samples2.writer_lock_held is False


def test_diff_command(repo_2_br_no_conf):
    runner = CliRunner()
    res = runner.invoke(cli.diff, ['master', 'testbranch'], obj=repo_2_br_no_conf)
    assert res.exit_code == 0


def test_commit_cli_message(repo_20_filled_samples2):
    co = repo_20_filled_samples2.checkout(write=True)
    co.add_str_column('test_meta')
    base_digest = co.commit_hash
    base_branch = co.branch_name
    co.close()
    assert base_branch == 'master'

    runner = CliRunner()
    res = runner.invoke(cli.commit, ['-m', 'this is my commit message'], obj=repo_20_filled_samples2)
    assert res.exit_code == 0
    out = res.stdout
    assert out.startswith('Commit message:\nthis is my commit message\nCommit Successful') is True
    new_digest = out.split(' ')[-1].rstrip('\n')
    assert new_digest != base_digest

    nco = repo_20_filled_samples2.checkout(write=True)
    try:
        assert nco.commit_hash == new_digest
        assert nco.branch_name == base_branch
    finally:
        nco.close()


def test_commit_cli_message_with_no_changes(repo_20_filled_samples2):
    co = repo_20_filled_samples2.checkout(write=True)
    base_digest = co.commit_hash
    base_branch = co.branch_name
    co.close()
    assert base_branch == 'master'

    runner = CliRunner()
    res = runner.invoke(cli.commit, ['-m', 'this is my commit message'], obj=repo_20_filled_samples2)
    assert res.exit_code == 1
    assert res.stdout.endswith('Error: No changes made in staging area. Cannot commit.\n')

    co = repo_20_filled_samples2.checkout(write=True)
    try:
        assert co.branch_name == base_branch
        assert co.commit_hash == base_digest
    finally:
        co.close()


def substitute_editor_commit_message(hint):
    return 'this is my commit message\n' + hint


def test_commit_editor_message(monkeypatch, repo_20_filled_samples2):
    import click
    monkeypatch.setattr(click, 'edit', substitute_editor_commit_message)

    co = repo_20_filled_samples2.checkout(write=True)
    co.add_str_column('test_meta')
    base_digest = co.commit_hash
    base_branch = co.branch_name
    co.close()
    assert base_branch == 'master'

    runner = CliRunner()
    res = runner.invoke(cli.commit, obj=repo_20_filled_samples2)
    assert res.exit_code == 0
    out = res.stdout
    assert out.startswith('Commit message:\nthis is my commit message\nCommit Successful') is True
    new_digest = out.split(' ')[-1].rstrip('\n')
    assert new_digest != base_digest

    nco = repo_20_filled_samples2.checkout(write=True)
    try:
        assert nco.commit_hash == new_digest
        assert nco.branch_name == base_branch
    finally:
        nco.close()


def substitute_editor_empty_commit_message(hint):
    return hint


def test_commit_editor_empty_message(monkeypatch, repo_20_filled_samples2):
    import click
    monkeypatch.setattr(click, 'edit', substitute_editor_empty_commit_message)

    co = repo_20_filled_samples2.checkout(write=True)
    co.add_str_column('test_meta')
    base_digest = co.commit_hash
    base_branch = co.branch_name
    co.close()
    assert base_branch == 'master'

    runner = CliRunner()
    res = runner.invoke(cli.commit, obj=repo_20_filled_samples2)
    assert res.exit_code == 0
    assert res.stdout == 'Aborted! Empty commit message\n'
    nco = repo_20_filled_samples2.checkout(write=True)
    try:
        assert nco.commit_hash == base_digest
        assert nco.branch_name == base_branch
    finally:
        nco.close()


def test_clone(written_two_cmt_server_repo):
    server, base_repo = written_two_cmt_server_repo
    runner = CliRunner()
    with runner.isolated_filesystem():
        P = getcwd()
        try:
            new_repo = Repository(P, exists=False)
            res = runner.invoke(
                cli.clone,
                ['--name', 'Foo Tester', '--email', 'foo@email.com', f'{server}'], obj=new_repo)

            assert res.exit_code == 0

            newLog = new_repo.log(return_contents=True)
            baseLog = base_repo.log(return_contents=True)
            assert newLog == baseLog
            assert new_repo.summary() == base_repo.summary()
        finally:
            new_repo._env._close_environments()


@pytest.mark.parametrize('backend', fixed_shape_backend_params)
def test_push_fetch_records(server_instance, backend):

    runner = CliRunner()
    with runner.isolated_filesystem():
        repo = Repository(getcwd(), exists=False)
        try:
            repo.init('foo', 'bar')
            dummyData = np.arange(50)
            co1 = repo.checkout(write=True, branch='master')
            co1.add_ndarray_column(name='dummy', prototype=dummyData, backend=backend)
            for idx in range(10):
                dummyData[:] = idx
                co1.columns['dummy'][str(idx)] = dummyData
            cmt1 = co1.commit('first commit adding dummy data')
            co1.close()

            repo.create_branch('testbranch')
            co2 = repo.checkout(write=True, branch='testbranch')
            for idx in range(10, 20):
                dummyData[:] = idx
                co2.columns['dummy'][str(idx)] = dummyData
            cmt2 = co2.commit('first commit on test branch adding non-conflict data')
            co2.close()

            repo.remote.add('origin', server_instance)

            res = runner.invoke(cli.push, ['origin', 'master'], obj=repo)
            assert res.exit_code == 0
            res = runner.invoke(cli.push, ['origin', 'testbranch'], obj=repo)
            assert res.exit_code == 0
        finally:
            repo._env._close_environments()


@pytest.mark.parametrize('backend', fixed_shape_backend_params)
@pytest.mark.parametrize('options', [
    ['origin', 'testbranch'],
    ['origin', 'master'],
    ['origin', 'testbranch', '--all-history'],
    ['origin', 'master', '--all-history'],
    ['origin', 'testbranch', '--column', 'data'],
    ['origin', 'master', '--column', 'data'],
    ['origin', 'testbranch', '--column', 'data', '--all-history'],
    ['origin', 'master', '--column', 'data', '--all-history'],
    ['origin', 'testbranch', '--column', 'data', '--all-history'],
    ['origin', 'master', '--nbytes', '3Kb'],
])
def test_fetch_records_and_data(server_instance, backend, options):
    runner = CliRunner()
    with runner.isolated_filesystem():
        repo = Repository(getcwd(), exists=False)
        try:
            repo.init('foo', 'bar')
            dummyData = np.arange(50)
            co1 = repo.checkout(write=True, branch='master')
            co1.add_ndarray_column(name='dummy', prototype=dummyData, backend=backend)
            for idx in range(10):
                dummyData[:] = idx
                co1.columns['dummy'][str(idx)] = dummyData
            cmt1 = co1.commit('first commit adding dummy data')
            co1.close()

            repo.create_branch('testbranch')
            co2 = repo.checkout(write=True, branch='testbranch')
            for idx in range(10, 20):
                dummyData[:] = idx
                co2.columns['dummy'][str(idx)] = dummyData
            cmt2 = co2.commit('first commit on test branch adding non-conflict data')
            co2.close()

            repo.remote.add('origin', server_instance)

            res = runner.invoke(cli.push, ['origin', 'master'], obj=repo)
            assert res.exit_code == 0
            res = runner.invoke(cli.push, ['origin', 'testbranch'], obj=repo)
            assert res.exit_code == 0
        finally:
            repo._env._close_environments()

    with runner.isolated_filesystem():
        repo = Repository(getcwd(), exists=False)
        try:
            res = runner.invoke(
                cli.clone,
                ['--name', 'Foo Tester', '--email', 'foo@email.com', f'{server_instance}'], obj=repo)
            assert res.exit_code == 0

            res = runner.invoke(cli.fetch_records, ['origin', 'testbranch'], obj=repo)
            assert res.exit_code == 0
            res = runner.invoke(cli.branch_create, ['testbranch', 'origin/testbranch'], obj=repo)
            assert res.exit_code == 0
            res = runner.invoke(cli.fetch_data, options, obj=repo)
            assert res.exit_code == 0
        finally:
            repo._env._close_environments()


def test_add_remote(managed_tmpdir):
    from hangar.remotes import RemoteInfo

    runner = CliRunner()
    with runner.isolated_filesystem():
        P = getcwd()
        repo = Repository(P, exists=False)
        try:
            res = runner.invoke(cli.init, ['--name', 'test', '--email', 'test@foo.com'], obj=repo)
            assert res.exit_code == 0

            res = runner.invoke(cli.add_remote, ['origin', 'localhost:50051'], obj=repo)
            assert res.exit_code == 0
            assert res.stdout == "RemoteInfo(name='origin', address='localhost:50051')\n"

            remote_list = repo.remote.list_all()
            assert remote_list == [RemoteInfo(name='origin', address='localhost:50051')]
        finally:
            repo._env._close_environments()


def test_remove_remote(managed_tmpdir):
    from hangar.remotes import RemoteInfo

    runner = CliRunner()
    with runner.isolated_filesystem():
        P = getcwd()
        repo = Repository(P, exists=False)
        try:
            res = runner.invoke(cli.init, ['--name', 'test', '--email', 'test@foo.com'], obj=repo)
            assert res.exit_code == 0

            res = runner.invoke(cli.add_remote, ['origin', 'localhost:50051'], obj=repo)
            assert res.exit_code == 0
            assert res.stdout == "RemoteInfo(name='origin', address='localhost:50051')\n"

            remote_list = repo.remote.list_all()
            assert remote_list == [RemoteInfo(name='origin', address='localhost:50051')]

            res = runner.invoke(cli.remove_remote, ['origin'], obj=repo)
            assert res.exit_code == 0
            assert res.stdout == "RemoteInfo(name='origin', address='localhost:50051')\n"
            assert repo.remote.list_all() == []
        finally:
            repo._env._close_environments()


def test_list_all_remotes(managed_tmpdir):
    from hangar.remotes import RemoteInfo

    runner = CliRunner()
    with runner.isolated_filesystem():
        P = getcwd()
        repo = Repository(P, exists=False)
        try:
            res = runner.invoke(cli.init, ['--name', 'test', '--email', 'test@foo.com'], obj=repo)
            assert res.exit_code == 0

            res = runner.invoke(cli.add_remote, ['origin', 'localhost:50051'], obj=repo)
            assert res.exit_code == 0
            assert res.stdout == "RemoteInfo(name='origin', address='localhost:50051')\n"
            res = runner.invoke(cli.add_remote, ['upstream', 'foo:ip'], obj=repo)
            assert res.exit_code == 0
            assert res.stdout == "RemoteInfo(name='upstream', address='foo:ip')\n"

            remote_list = repo.remote.list_all()
            assert remote_list == [
                RemoteInfo(name='origin', address='localhost:50051'),
                RemoteInfo(name='upstream', address='foo:ip')
            ]

            res = runner.invoke(cli.list_remotes, obj=repo)
            assert res.exit_code == 0
            expected_stdout = "[RemoteInfo(name='origin', address='localhost:50051'), "\
                              "RemoteInfo(name='upstream', address='foo:ip')]\n"
            assert res.stdout == expected_stdout
        finally:
            repo._env._close_environments()


def test_summary(written_two_cmt_server_repo, capsys):
    server, base_repo = written_two_cmt_server_repo
    runner = CliRunner()
    with runner.isolated_filesystem():
        try:
            with capsys.disabled():
                P = getcwd()
                new_repo = Repository(P, exists=False)
                res = runner.invoke(
                    cli.clone,
                    ['--name', 'Foo Tester', '--email', 'foo@email.com', f'{server}'], obj=new_repo)

                assert res.exit_code == 0
                assert new_repo.summary() == base_repo.summary()

            new_repo.summary()

            with capsys.disabled():
                res = runner.invoke(cli.summary, obj=new_repo)
                assert res.stdout == f"{capsys.readouterr().out}\n"
        finally:
            new_repo._env._close_environments


def test_log(written_two_cmt_server_repo, capsys):
    server, base_repo = written_two_cmt_server_repo
    runner = CliRunner()
    with runner.isolated_filesystem():
        try:
            with capsys.disabled():
                P = getcwd()
                new_repo = Repository(P, exists=False)
                res = runner.invoke(
                    cli.clone,
                    ['--name', 'Foo Tester', '--email', 'foo@email.com', f'{server}'], obj=new_repo)

                assert res.exit_code == 0
                assert new_repo.log() == base_repo.log()

            new_repo.log()

            with capsys.disabled():
                res = runner.invoke(cli.log, ['master'], obj=new_repo)
                assert res.stdout == f"{capsys.readouterr().out}\n"
        finally:
            new_repo._env._close_environments()


def test_status(repo_20_filled_samples2):
    from hangar.records.summarize import status
    repo = repo_20_filled_samples2

    dummyData = np.arange(50).astype(np.int64)
    co2 = repo.checkout(write=True)
    for idx in range(10, 20):
        dummyData[:] = idx
        co2.columns['dummy'][str(idx)] = dummyData
        co2.columns['dummy'][idx] = dummyData
    df = co2.diff.staged()
    co2.close()
    expected = status(repo._env.hashenv, 'master', df.diff).getvalue()
    runner = CliRunner()
    res = runner.invoke(cli.status, obj=repo)
    assert res.exit_code == 0
    assert res.stdout == expected


def test_arrayset_create_uint8(repo_20_filled_samples2):
    runner = CliRunner()
    res = runner.invoke(
        cli.create_column,
        ['train_images', 'UINT8', '256', '256', '3'], obj=repo_20_filled_samples2)
    assert res.exit_code == 0
    assert res.stdout == 'Initialized Column: train_images\n'
    co = repo_20_filled_samples2.checkout(write=True)
    try:
        assert 'train_images' in co.columns
        assert co.columns['train_images'].shape == (256, 256, 3)
        assert co.columns['train_images'].dtype == np.uint8
        assert co.columns['train_images'].schema_type == 'fixed_shape'
        assert len(co.columns['train_images']) == 0
    finally:
        co.close()


def test_arrayset_create_float32(repo_20_filled_samples2):
    runner = CliRunner()
    res = runner.invoke(
        cli.create_column,
        ['train_images', 'FLOAT32', '256'], obj=repo_20_filled_samples2)
    assert res.exit_code == 0
    assert res.stdout == 'Initialized Column: train_images\n'
    co = repo_20_filled_samples2.checkout(write=True)
    try:
        assert 'train_images' in co.columns
        assert co.columns['train_images'].shape == (256,)
        assert co.columns['train_images'].dtype == np.float32
        assert co.columns['train_images'].schema_type == 'fixed_shape'
        assert len(co.columns['train_images']) == 0
    finally:
        co.close()


def test_arrayset_create_invalid_dtype_fails(repo_20_filled_samples2):
    runner = CliRunner()
    res = runner.invoke(
        cli.create_column,
        ['train_images', 'FLOAT7', '256'], obj=repo_20_filled_samples2)
    assert res.exit_code == 2
    expected = ('invalid choice: FLOAT7. (choose from UINT8, '
                'INT8, UINT16, INT16, UINT32, INT32, UINT64, '
                'INT64, FLOAT16, FLOAT32, FLOAT64, STR)\n')
    assert res.stdout.endswith(expected) is True
    co = repo_20_filled_samples2.checkout(write=True)
    try:
        assert 'train_images' not in co.columns
    finally:
        co.close()


def test_arrayset_create_invalid_name_fails(repo_20_filled_samples2):
    runner = CliRunner()
    res = runner.invoke(cli.create_column, ['tra#in', 'FLOAT32', '256'], obj=repo_20_filled_samples2)
    assert res.exit_code == 1
    msg = res.stdout
    assert msg.startswith('Error: Column name provided: `tra#in` is invalid.') is True
    co = repo_20_filled_samples2.checkout(write=True)
    try:
        assert 'tra#in' not in co.columns
        assert 'dummy' in co.columns
        assert len(co.columns) == 1
    finally:
        co.close()


def test_arrayset_create_variable_shape(repo_20_filled_samples2):
    runner = CliRunner()
    res = runner.invoke(
        cli.create_column,
        ['train_images', 'FLOAT32', '256', '--variable-shape'], obj=repo_20_filled_samples2)
    assert res.exit_code == 0
    assert res.stdout == 'Initialized Column: train_images\n'
    co = repo_20_filled_samples2.checkout(write=True)
    try:
        assert 'train_images' in co.columns
        assert co.columns['train_images'].shape == (256,)
        assert co.columns['train_images'].dtype == np.float32
        assert co.columns['train_images'].schema_type == 'variable_shape'
        assert co.columns['train_images'].contains_subsamples is False
        assert len(co.columns['train_images']) == 0
    finally:
        co.close()


def test_arrayset_create_contains_subsamples(repo_20_filled_samples2):
    runner = CliRunner()
    res = runner.invoke(
        cli.create_column,
        ['train_images', 'FLOAT32', '256', '--contains-subsamples'], obj=repo_20_filled_samples2)
    assert res.exit_code == 0
    assert res.stdout == 'Initialized Column: train_images\n'
    co = repo_20_filled_samples2.checkout(write=True)
    try:
        assert 'train_images' in co.columns
        assert co.columns['train_images'].shape == (256,)
        assert co.columns['train_images'].dtype == np.float32
        assert co.columns['train_images'].schema_type == 'fixed_shape'
        assert co.columns['train_images'].contains_subsamples is True
        assert len(co.columns['train_images']) == 0
    finally:
        co.close()


def test_remove_arrayset(repo_20_filled_samples2):
    runner = CliRunner()
    res = runner.invoke(cli.remove_column, ['dummy'], obj=repo_20_filled_samples2)
    assert res.exit_code == 0
    assert res.stdout == 'Successfully removed column: dummy\n'
    co = repo_20_filled_samples2.checkout(write=True)
    try:
        assert 'repo_20_filled_samples2' not in co.columns
        assert len(co.columns) == 0
    finally:
        co.close()


def test_remove_non_existing_arrayset(repo_20_filled_samples2):
    runner = CliRunner()
    res = runner.invoke(cli.remove_column, ['doesnotexist'], obj=repo_20_filled_samples2)
    assert res.exit_code == 1
    assert res.stdout == "Error: 'Cannot remove: doesnotexist. Key does not exist.'\n"
    co = repo_20_filled_samples2.checkout(write=True)
    try:
        assert 'doesnotexist' not in co.columns
        assert 'dummy' in co.columns
        assert len(co.columns) == 1
        assert len(co.columns['dummy']) == 10
    finally:
        co.close()


def test_branch_create_and_list(written_two_cmt_server_repo):
    server, base_repo = written_two_cmt_server_repo

    co = base_repo.checkout(write=True)
    cmt = co.commit_hash
    co.close()

    runner = CliRunner()
    with runner.isolated_filesystem():
        P = getcwd()
        new_repo = Repository(P, exists=False)
        try:
            res = runner.invoke(
                cli.clone,
                ['--name', 'Foo Tester', '--email', 'foo@email.com', f'{server}'], obj=new_repo)
            assert res.exit_code == 0

            res = runner.invoke(cli.branch_create, ['testbranch'], obj=new_repo)
            assert res.exit_code == 0
            assert res.stdout == f"Created BRANCH: testbranch HEAD: {cmt}\n"

            branches = new_repo.list_branches()
            assert branches == ['master', 'origin/master', 'testbranch']

            res = runner.invoke(cli.branch_list, obj=new_repo)
            assert res.exit_code == 0
            assert res.stdout == "['master', 'origin/master', 'testbranch']\n"
        finally:
            new_repo._env._close_environments()


@pytest.mark.filterwarnings("ignore:Column.* contains `reference-only` samples")
def test_branch_create_and_delete(written_two_cmt_server_repo):
    server, base_repo = written_two_cmt_server_repo

    co = base_repo.checkout(write=True)
    cmt = co.commit_hash
    co.close()

    runner = CliRunner()
    with runner.isolated_filesystem():
        P = getcwd()
        new_repo = Repository(P, exists=False)
        try:
            res = runner.invoke(
                cli.clone,
                ['--name', 'Foo Tester', '--email', 'foo@email.com', f'{server}'], obj=new_repo)
            assert res.exit_code == 0

            res = runner.invoke(cli.branch_create, ['testbranch'], obj=new_repo)
            assert res.exit_code == 0
            assert res.stdout == f"Created BRANCH: testbranch HEAD: {cmt}\n"

            branches = new_repo.list_branches()
            assert branches == ['master', 'origin/master', 'testbranch']

            res = runner.invoke(cli.branch_remove, ['testbranch'], obj=new_repo)
            assert res.exit_code == 0
            assert res.stdout == f"Deleted BRANCH: testbranch HEAD: {cmt}\n"

            branches = new_repo.list_branches()
            assert branches == ['master', 'origin/master']

            new_repo.create_branch('secondtest')
            co = new_repo.checkout(write=True, branch='secondtest')
            co.add_str_column('test_meta')
            newDigest = co.commit('dummy commit')
            co.close()

            # re-open with staging set to master so we can try to delete secondtest
            co = new_repo.checkout(write=True, branch='master')
            co.close()

            res = runner.invoke(cli.branch_remove, ['secondtest'], obj=new_repo)
            assert res.exit_code == 1

            res = runner.invoke(cli.branch_remove, ['secondtest', '-f'], obj=new_repo)
            assert res.exit_code == 0
            assert res.stdout == f"Deleted BRANCH: secondtest HEAD: {newDigest}\n"

            res = runner.invoke(cli.branch_list, obj=new_repo)
            assert res.exit_code == 0
            assert res.stdout == "['master', 'origin/master']\n"
        finally:
            new_repo._env._close_environments()


def test_start_server(managed_tmpdir):
    import time
    runner = CliRunner()
    with runner.isolated_filesystem():
        startTime = time.time()
        res = runner.invoke(cli.server, ['--ip', 'localhost', '--port', '50111', '--timeout', '1'])
        assert time.time() - startTime <= 1.8  # buffer to give it time to stop
        assert res.exit_code == 0
        assert 'Hangar Server Started' in res.stdout


# ------------------------ Developer Commands --------------------------------


def test_db_view_command(repo_20_filled_samples):
    repo = repo_20_filled_samples
    runner = CliRunner()
    res = runner.invoke(cli.lmdb_record_details, ['-a'], obj=repo)

    dbs_queried = 0
    assert res.exit_code == 0
    for line in res.stdout.splitlines():
        if '.lmdb' in line:
            dbs_queried += 1
    assert dbs_queried == 5

    res = runner.invoke(cli.lmdb_record_details, ['-a', '--limit', '10'], obj=repo)
    assert res.exit_code == 0


# =========================== External Plugin =================================


def monkeypatch_scan(provides, accepts, attribute, func):
    def wrapper(self):
        from hangar.external import BasePlugin

        plugin = BasePlugin(provides, accepts)
        plugin.__dict__[attribute] = func

        self._plugin_store['myplugin'] = plugin
    return wrapper


@pytest.fixture()
def written_repo_with_1_sample(aset_samples_initialized_repo):
    aset_name = 'writtenaset'
    shape = (5, 7)
    co = aset_samples_initialized_repo.checkout(write=True)
    aset = co.columns[aset_name]
    aset['data'] = np.random.random(shape)
    aset['123'] = np.random.random(shape)
    aset[123] = np.random.random(shape)
    co.commit('added')
    co.close()
    yield aset_samples_initialized_repo


class TestImport(object):

    @staticmethod
    def load(fpath, *args, **kwargs):
        data = np.random.random((5, 7)).astype(np.float64)
        if isinstance(fpath, Path):
            fpath = fpath.name
        return data, fpath

    def test_import(self, monkeypatch, written_repo_with_1_sample):
        repo = written_repo_with_1_sample
        runner = CliRunner()
        shape = (5, 7)
        fpath = 'data.ext'
        aset_name = 'writtenaset'

        with monkeypatch.context() as m, runner.isolated_filesystem():
            with open('data.ext', 'w') as f:
                f.write('test')

            m.setattr(PluginManager, "_scan_plugins", monkeypatch_scan(['load'], ['ext'], 'load', self.load))
            # adding data
            res = runner.invoke(cli.import_data, [aset_name, fpath], obj=repo)
            assert res.exit_code == 0
            co = repo.checkout(write=True)
            co.commit('added data')
            d1 = co.columns[aset_name][fpath]
            co.close()

            # without overwrite
            res = runner.invoke(cli.import_data, [aset_name, fpath], obj=repo)
            assert res.exit_code == 0
            co = repo.checkout()
            d2 = co.columns[aset_name][fpath]
            co.close()
            assert np.allclose(d1, d2)

            # with overwrite
            res = runner.invoke(cli.import_data, [aset_name, fpath, '--overwrite'], obj=repo)
            assert res.exit_code == 0
            co = repo.checkout(write=True)
            co.commit('added data')
            d3 = co.columns[aset_name][fpath]
            co.close()
            assert not np.allclose(d1, d3)
        assert d1.shape == d2.shape == d3.shape == shape

    def test_import_wrong_args(self, monkeypatch, written_repo_with_1_sample):
        repo = written_repo_with_1_sample
        runner = CliRunner()

        aset_name = 'writtenaset'

        with monkeypatch.context() as m:
            m.setattr(PluginManager, "_scan_plugins", monkeypatch_scan(['load'], ['ext'], 'load', self.load))

            with runner.isolated_filesystem():

                # invalid file
                res = runner.invoke(cli.import_data, [aset_name, 'valid.ext'], obj=repo)
                assert res.exit_code == 2
                assert "Invalid value for" in res.stdout
                assert "PATH" in res.stdout
                assert "valid.ext" in res.stdout
                assert "does not exist." in res.stdout

                with open('valid.ext', 'w') as f:
                    f.write('empty')

                with open('valid.ext.bz2', 'w') as f:
                    f.write('empty')

                res = runner.invoke(cli.import_data, [aset_name, 'valid.ext.bz2'], obj=repo)
                assert res.exit_code == 1
                assert res.stdout.endswith('No plugins found for the file extension ext.bz2 that could do load\n')

                # invalid branch
                res = runner.invoke(cli.import_data, [aset_name, 'valid.ext', '--branch', 'invalid'], obj=repo)
                assert res.exit_code == 1
                assert res.stdout.endswith('Branch name: invalid does not exist, Exiting.\n')

                # invalid plugin
                res = runner.invoke(cli.import_data, [aset_name, 'valid.ext', '--plugin', 'invalid'], obj=repo)
                assert res.exit_code == 1
                assert res.stdout.endswith('Plugin invalid not found\n')

    def test_import_generator_on_load(self, monkeypatch, written_repo_with_1_sample):
        repo = written_repo_with_1_sample
        runner = CliRunner()
        fpath = 'data.ext'
        aset_name = 'writtenaset'

        def load(fpath, *args, **kwargs):
            for i in range(10):
                data, name = self.load(fpath, *args, **kwargs)
                if isinstance(name, Path):
                    name = name.name
                yield data, f"{i}_{name}"

        with monkeypatch.context() as m, runner.isolated_filesystem():
            with open('data.ext', 'w') as f:
                f.write('test')
            m.setattr(PluginManager, "_scan_plugins", monkeypatch_scan(['load'], ['ext'], 'load', load))
            res = runner.invoke(cli.import_data, [aset_name, fpath], obj=repo)
            assert res.exit_code == 0
            co = repo.checkout(write=True)
            co.commit('added data')
            aset = co.columns[aset_name]
            for i in range(10):
                assert f"{i}_{fpath}" in aset.keys()
            co.close()


class TestExport(object):
    save_msg = "Data saved from custom save function"

    @classmethod
    def save(cls, data, outdir, sampleN, extension, *args, **kwargs):
        print(cls.save_msg)
        fpath = os.path.join(outdir, f"{sampleN}.{extension}")
        print(fpath)

    def test_export_success(self, monkeypatch, written_repo_with_1_sample, tmp_path):
        repo = written_repo_with_1_sample
        runner = CliRunner()
        aset_name = 'writtenaset'

        with monkeypatch.context() as m:
            m.setattr(PluginManager, "_scan_plugins", monkeypatch_scan(['save'], ['ext'], 'save', self.save))

            # single sample
            res = runner.invoke(
                cli.export_data, [aset_name, '-o', str(tmp_path), '--sample', 'data', '--format', 'ext'], obj=repo)
            assert res.exit_code == 0
            assert self.save_msg in res.output

            # with sample name and sample type
            res = runner.invoke(
                cli.export_data, [aset_name, '-o', str(tmp_path), '--sample', 'int:123', '--format', 'ext'], obj=repo)
            assert res.exit_code == 0
            assert os.path.join(tmp_path, 'int:123.ext') in res.output
            res = runner.invoke(
                cli.export_data, [aset_name, '-o', str(tmp_path), '--sample', 'str:123', '--format', 'ext'], obj=repo)
            assert res.exit_code == 0
            assert os.path.join(tmp_path, 'str:123.ext') in res.output
            res = runner.invoke(
                cli.export_data, [aset_name, '-o', str(tmp_path), '--sample', '123', '--format', 'ext'], obj=repo)
            assert res.exit_code == 0
            assert os.path.join(tmp_path, 'str:123.ext') in res.output

            # whole column
            res = runner.invoke(
                cli.export_data, [aset_name, '-o', str(tmp_path), '--format', 'ext'], obj=repo)
            assert res.exit_code == 0
            assert os.path.join(tmp_path, 'str:data.ext') in res.output
            assert os.path.join(tmp_path, 'str:123.ext') in res.output
            assert os.path.join(tmp_path, 'int:123.ext') in res.output

    def test_export_wrong_out_location(self, monkeypatch, written_repo_with_1_sample):
        repo = written_repo_with_1_sample
        runner = CliRunner()
        aset_name = 'writtenaset'

        with monkeypatch.context() as m:
            m.setattr(PluginManager, "_scan_plugins", monkeypatch_scan(['save'], ['ext'], 'save', self.save))

            # single sample
            res = runner.invoke(
                cli.export_data, [aset_name, '-o', 'wrongpath', '--sample', 'data', '--format', 'ext'], obj=repo)
            assert res.exit_code == 2
            assert "Invalid value for" in res.stdout
            assert "-o" in res.stdout
            assert "--out" in res.stdout

    def test_export_wrong_arg(self, monkeypatch, written_repo_with_1_sample, tmp_path):
        repo = written_repo_with_1_sample
        runner = CliRunner()
        aset_name = 'writtenaset'

        with monkeypatch.context() as m:
            m.setattr(PluginManager, "_scan_plugins", monkeypatch_scan(['save'], ['ext'], 'save', self.save))
            res = runner.invoke(
                cli.export_data, [aset_name, '-o', str(tmp_path), '--plugin', 'invalid'], obj=repo)
            assert res.exit_code == 1
            assert 'Plugin invalid not found' in res.stdout

    def test_export_without_specifying_out(self, monkeypatch, written_repo_with_1_sample):
        import os
        repo = written_repo_with_1_sample
        runner = CliRunner()
        aset_name = 'writtenaset'

        with monkeypatch.context() as m:
            m.setattr(PluginManager, "_scan_plugins", monkeypatch_scan(['save'], ['ext'], 'save', self.save))
            res = runner.invoke(
                cli.export_data, [aset_name, '--sample', 'data', '--format', 'ext'], obj=repo)
            assert os.getcwd() in res.output

    def test_export_for_non_existent_sample(self, monkeypatch, written_repo_with_1_sample):
        repo = written_repo_with_1_sample
        runner = CliRunner()
        aset_name = 'writtenaset'

        with monkeypatch.context() as m:
            m.setattr(PluginManager, "_scan_plugins", monkeypatch_scan(['save'], ['ext'], 'save', self.save))
            res = runner.invoke(
                cli.export_data, [aset_name, '--sample', 'wrongname', '--format', 'ext'], obj=repo)
            assert res.exit_code == 1
            assert 'wrongname' in res.output

    def test_export_for_specified_branch(self, monkeypatch, written_repo_with_1_sample):
        repo = written_repo_with_1_sample
        runner = CliRunner()
        aset_name = 'writtenaset'

        with monkeypatch.context() as m:
            m.setattr(PluginManager, "_scan_plugins", monkeypatch_scan(['save'], ['ext'], 'save', self.save))
            res = runner.invoke(
                cli.export_data, [aset_name, 'master', '--sample', 'data', '--format', 'ext'], obj=repo)
            assert res.exit_code == 0


class TestShow(object):
    show_msg = "Data is displayed from custom show function"

    @classmethod
    def show(cls, fpath, *args, **kwargs):
        print(cls.show_msg)

    def test_show_success(self, monkeypatch, written_repo_with_1_sample):
        repo = written_repo_with_1_sample
        runner = CliRunner()
        aset_name = 'writtenaset'

        with monkeypatch.context() as m:
            m.setattr(PluginManager, "_scan_plugins", monkeypatch_scan(['show'], ['ext'], 'show', self.show))
            res = runner.invoke(
                cli.view_data, [aset_name, 'data', '--format', 'ext'], obj=repo)
            assert res.exit_code == 0
            assert self.show_msg in res.output

    def test_show_on_startpoint(self, monkeypatch, written_repo_with_1_sample):
        repo = written_repo_with_1_sample
        runner = CliRunner()
        aset_name = 'writtenaset'

        with monkeypatch.context() as m:
            m.setattr(PluginManager, "_scan_plugins", monkeypatch_scan(['show'], ['ext'], 'show', self.show))
            res = runner.invoke(
                cli.view_data, [aset_name, 'data', 'master', '--format', 'ext'], obj=repo)
            assert res.exit_code == 0
            res = runner.invoke(
                cli.view_data, [aset_name, 'data', 'wrongstartpoint', '--format', 'ext'], obj=repo)
            assert "No matching commit hash found" in str(res.exception)

    def test_show_with_wrong_arg(self, monkeypatch, written_repo_with_1_sample):
        repo = written_repo_with_1_sample
        runner = CliRunner()
        aset_name = 'writtenaset'

        with monkeypatch.context() as m:
            m.setattr(PluginManager, "_scan_plugins", monkeypatch_scan(['show'], ['ext'], 'show', self.show))
            res = runner.invoke(
                cli.view_data, [aset_name, 'data', '--format', 'wrong'], obj=repo)
            assert res.exit_code == 1
            assert 'No plugins found' in res.stdout

    def test_wrong_sample_name(self, monkeypatch, written_repo_with_1_sample):
        repo = written_repo_with_1_sample
        runner = CliRunner()
        aset_name = 'writtenaset'
        with monkeypatch.context() as m:
            m.setattr(PluginManager, "_scan_plugins", monkeypatch_scan(['show'], ['ext'], 'show', self.show))
            res = runner.invoke(
                cli.view_data, [aset_name, 'wrongsample', '--format', 'ext'], obj=repo)
            assert res.exit_code == 1
            assert "wrongsample" in res.stdout



import os
from pathlib import Path
from typing import Union
import tempfile
import warnings
from concurrent import futures
from os.path import join as pjoin
import shutil
import configparser

import blosc
import grpc
import lmdb

from . import chunks
from . import hangar_service_pb2
from . import hangar_service_pb2_grpc
from . import request_header_validator_interceptor
from .content import ContentWriter
from .. import constants as c
from ..context import Environments
from ..txnctx import TxnRegister
from ..backends import BACKEND_ACCESSOR_MAP, backend_decoder
from ..records import commiting, hashs, heads, parsing, queries, summarize
from ..records import (
    hash_schema_db_key_from_raw_key,
    hash_data_db_key_from_raw_key,
)
from ..records.hashmachine import ndarray_hasher_tcode_0
from ..utils import set_blosc_nthreads

set_blosc_nthreads()


class HangarServer(hangar_service_pb2_grpc.HangarServiceServicer):

    def __init__(self, repo_path: Union[str, bytes, Path], overwrite=False):

        if isinstance(repo_path, (str, bytes)):
            repo_path = Path(repo_path)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            envs = Environments(pth=repo_path)
        self.env: Environments = envs

        try:
            self.env.init_repo(
                user_name='SERVER_USER',
                user_email='SERVER_USER@HANGAR.SERVER',
                remove_old=overwrite)
        except OSError:
            pass

        self._rFs = {}
        for backend, accessor in BACKEND_ACCESSOR_MAP.items():
            if accessor is not None:
                self._rFs[backend] = accessor(
                    repo_path=self.env.repo_path,
                    schema_shape=None,
                    schema_dtype=None)
                self._rFs[backend].open(mode='r')

        src_path = pjoin(os.path.dirname(__file__), c.CONFIG_SERVER_NAME)
        dst_path = pjoin(repo_path, c.CONFIG_SERVER_NAME)
        if not os.path.isfile(dst_path):
            shutil.copyfile(src_path, dst_path)
        self.CFG = configparser.ConfigParser()
        self.CFG.read(dst_path)

        self.txnregister = TxnRegister()
        self.repo_path = self.env.repo_path
        self.data_dir = pjoin(self.repo_path, c.DIR_DATA)
        self.CW = ContentWriter(self.env)

    def close(self):
        for backend_accessor in self._rFs.values():
            backend_accessor.close()
        self.env._close_environments()

    # -------------------- Client Config --------------------------------------

    def PING(self, request, context):
        """Test function. PING -> PONG!
        """
        reply = hangar_service_pb2.PingReply(result='PONG')
        return reply

    def GetClientConfig(self, request, context):
        """Return parameters to the client to set up channel options as desired by the server.
        """
        clientCFG = self.CFG['CLIENT_GRPC']
        push_max_nbytes = clientCFG['push_max_nbytes']
        enable_compression = clientCFG['enable_compression']
        optimization_target = clientCFG['optimization_target']

        err = hangar_service_pb2.ErrorProto(code=0, message='OK')
        reply = hangar_service_pb2.GetClientConfigReply(error=err)
        reply.config['push_max_nbytes'] = push_max_nbytes
        reply.config['enable_compression'] = enable_compression
        reply.config['optimization_target'] = optimization_target
        return reply

    # -------------------- Branch Record --------------------------------------

    def FetchBranchRecord(self, request, context):
        """Return the current HEAD commit of a particular branch
        """
        branch_name = request.rec.name
        try:
            head = heads.get_branch_head_commit(self.env.branchenv, branch_name)
            rec = hangar_service_pb2.BranchRecord(name=branch_name, commit=head)
            err = hangar_service_pb2.ErrorProto(code=0, message='OK')
            reply = hangar_service_pb2.FetchBranchRecordReply(rec=rec, error=err)
        except ValueError:
            msg = f'BRANCH: {branch_name} DOES NOT EXIST ON SERVER.'
            context.set_details(msg)
            context.set_code(grpc.StatusCode.NOT_FOUND)
            err = hangar_service_pb2.ErrorProto(code=5, message=msg)
            reply = hangar_service_pb2.FetchBranchRecordReply(error=err)
        return reply

    def PushBranchRecord(self, request, context):
        """Update the HEAD commit of a branch, creating the record if not previously existing.
        """
        branch_name = request.rec.name
        commit = request.rec.commit
        branch_names = heads.get_branch_names(self.env.branchenv)
        if branch_name not in branch_names:
            heads.create_branch(self.env.branchenv, name=branch_name, base_commit=commit)
            err = hangar_service_pb2.ErrorProto(code=0, message='OK')
        else:
            current_head = heads.get_branch_head_commit(self.env.branchenv, branch_name)
            if current_head == commit:
                msg = f'NO CHANGE TO BRANCH: {branch_name} WITH HEAD: {current_head}'
                context.set_details(msg)
                context.set_code(grpc.StatusCode.ALREADY_EXISTS)
                err = hangar_service_pb2.ErrorProto(code=6, message=msg)
            else:
                heads.set_branch_head_commit(self.env.branchenv, branch_name, commit)
                err = hangar_service_pb2.ErrorProto(code=0, message='OK')

        reply = hangar_service_pb2.PushBranchRecordReply(error=err)
        return reply

    # -------------------------- Commit Record --------------------------------

    def FetchCommit(self, request, context):
        """Return raw data representing contents, spec, and parents of a commit hash.
        """
        commit = request.commit
        commitRefKey = parsing.commit_ref_db_key_from_raw_key(commit)
        commitParentKey = parsing.commit_parent_db_key_from_raw_key(commit)
        commitSpecKey = parsing.commit_spec_db_key_from_raw_key(commit)

        reftxn = self.txnregister.begin_reader_txn(self.env.refenv)
        try:
            commitRefVal = reftxn.get(commitRefKey, default=False)
            commitParentVal = reftxn.get(commitParentKey, default=False)
            commitSpecVal = reftxn.get(commitSpecKey, default=False)
        finally:
            self.txnregister.abort_reader_txn(self.env.refenv)

        if commitRefVal is False:
            msg = f'COMMIT: {commit} DOES NOT EXIST ON SERVER'
            context.set_details(msg)
            context.set_code(grpc.StatusCode.NOT_FOUND)
            err = hangar_service_pb2.ErrorProto(code=5, message=msg)
            reply = hangar_service_pb2.FetchCommitReply(commit=commit, error=err)
            yield reply
            raise StopIteration()
        else:
            raw_data_chunks = chunks.chunk_bytes(commitRefVal)
            bsize = len(commitRefVal)
            commit_proto = hangar_service_pb2.CommitRecord()
            commit_proto.parent = commitParentVal
            commit_proto.spec = commitSpecVal
            reply = hangar_service_pb2.FetchCommitReply(commit=commit, total_byte_size=bsize)
            for chunk in raw_data_chunks:
                commit_proto.ref = chunk
                reply.record.CopyFrom(commit_proto)
                yield reply

    def PushCommit(self, request_iterator, context):
        """Record the contents of a new commit sent to the server.

        Will not overwrite data if a commit hash is already recorded on the server.
        """
        for idx, request in enumerate(request_iterator):
            if idx == 0:
                commit = request.commit
                refBytes, offset = bytearray(request.total_byte_size), 0
                specVal = request.record.spec
                parentVal = request.record.parent
            size = len(request.record.ref)
            refBytes[offset: offset + size] = request.record.ref
            offset += size

        digest = self.CW.commit(commit, parentVal, specVal, refBytes)
        if not digest:
            msg = f'COMMIT: {commit} ALREADY EXISTS'
            context.set_code(grpc.StatusCode.ALREADY_EXISTS)
            context.set_details(msg)
            err = hangar_service_pb2.ErrorProto(code=6, message=msg)
        else:
            err = hangar_service_pb2.ErrorProto(code=0, message='OK')
            commiting.move_process_data_to_store(self.env.repo_path, remote_operation=True)

        reply = hangar_service_pb2.PushCommitReply(error=err)
        return reply

    # --------------------- Schema Record -------------------------------------

    def FetchSchema(self, request, context):
        """Return the raw byte specification of a particular schema with requested hash.
        """
        schema_hash = request.rec.digest
        schemaKey = hash_schema_db_key_from_raw_key(schema_hash)
        hashTxn = self.txnregister.begin_reader_txn(self.env.hashenv)
        try:
            schemaExists = hashTxn.get(schemaKey, default=False)
            if schemaExists is not False:
                print(f'found schema: {schema_hash}')
                rec = hangar_service_pb2.SchemaRecord(digest=schema_hash, blob=schemaExists)
                err = hangar_service_pb2.ErrorProto(code=0, message='OK')
            else:
                print(f'not exists: {schema_hash}')
                msg = f'SCHEMA HASH: {schema_hash} DOES NOT EXIST ON SERVER'
                context.set_details(msg)
                context.set_code(grpc.StatusCode.NOT_FOUND)
                err = hangar_service_pb2.ErrorProto(code=5, message=msg)
                rec = hangar_service_pb2.SchemaRecord(digest=schema_hash)
        finally:
            self.txnregister.abort_reader_txn(self.env.hashenv)

        reply = hangar_service_pb2.FetchSchemaReply(rec=rec, error=err)
        return reply

    def PushSchema(self, request, context):
        """Add a new schema byte specification record.

        Will not overwrite a schema hash which already exists on the server.
        """
        schema_hash = request.rec.digest
        schema_val = request.rec.blob

        digest = self.CW.schema(schema_hash, schema_val)
        if not digest:
            print(f'exists: {schema_val}')
            msg = f'SCHEMA: {schema_hash} ALREADY EXISTS ON SERVER'
            context.set_details(msg)
            context.set_code(grpc.StatusCode.ALREADY_EXISTS)
            err = hangar_service_pb2.ErrorProto(code=6, message=msg)
        else:
            print(f'created new: {schema_val}')
            err = hangar_service_pb2.ErrorProto(code=0, message='OK')
        reply = hangar_service_pb2.PushSchemaReply(error=err)
        return reply

    # ---------------------------- Data ---------------------------------------

    def FetchData(self, request_iterator, context):
        """Return a packed byte representation of samples corresponding to a digest.

        Please see comments below which explain why not all requests are
        guaranteed to fully complete in one operation.

        We receive a list of digests to send to the client. One consideration
        we have is that there is no way to know how much memory will be used
        when the data is read from disk. Samples are compressed against
        each-other before going over the wire, which means its preferable to
        read in as much as possible. However, since we don't want to overload
        the client system when the binary blob is decompressed into individual
        tensors, we set some maximum size which tensors can occupy when
        uncompressed. When we receive a list of digests whose data size is in
        excess of this limit, we just say sorry to the client, send the chunk
        of digests/tensors off to them as is (incomplete), and request that
        the client figure out what it still needs and ask us again.
        """
        for idx, request in enumerate(request_iterator):
            if idx == 0:
                uncomp_nbytes = request.uncomp_nbytes
                comp_nbytes = request.comp_nbytes
                dBytes, offset = bytearray(comp_nbytes), 0
            size = len(request.raw_data)
            dBytes[offset: offset + size] = request.raw_data
            offset += size

        uncompBytes = blosc.decompress(dBytes)
        if uncomp_nbytes != len(uncompBytes):
            msg = f'Expected nbytes data sent: {uncomp_nbytes} != received {comp_nbytes}'
            context.set_details(msg)
            context.set_code(grpc.StatusCode.DATA_LOSS)
            err = hangar_service_pb2.ErrorProto(code=15, message=msg)
            reply = hangar_service_pb2.FetchDataReply(error=err)
            yield reply
            raise StopIteration()

        totalSize, records = 0, []
        unpacked_digests = uncompBytes.decode().split(c.SEP_LST)
        hashTxn = self.txnregister.begin_reader_txn(self.env.hashenv)

        try:
            fetch_max_nbytes = int(self.CFG['SERVER_GRPC']['fetch_max_nbytes'])
            for digest in unpacked_digests:
                hashKey = hash_data_db_key_from_raw_key(digest)
                hashVal = hashTxn.get(hashKey, default=False)
                if hashVal is False:
                    msg = f'HASH DOES NOT EXIST: {hashKey}'
                    context.set_details(msg)
                    context.set_code(grpc.StatusCode.NOT_FOUND)
                    err = hangar_service_pb2.ErrorProto(code=5, message=msg)
                    reply = hangar_service_pb2.FetchDataReply(error=err)
                    yield reply
                    raise StopIteration()
                else:
                    spec = backend_decoder(hashVal)
                    tensor = self._rFs[spec.backend].read_data(spec)

                record = chunks.serialize_record(tensor, digest, '')
                records.append(record)
                totalSize += len(record)
                if totalSize >= fetch_max_nbytes:
                    pack = chunks.serialize_record_pack(records)
                    err = hangar_service_pb2.ErrorProto(code=0, message='OK')
                    cIter = chunks.tensorChunkedIterator(
                        buf=pack, uncomp_nbytes=len(pack), itemsize=tensor.itemsize,
                        pb2_request=hangar_service_pb2.FetchDataReply, err=err)
                    yield from cIter
                    msg = 'HANGAR REQUESTED RETRY: developer enforced limit on returned '\
                          'raw data size to prevent memory overload of user system.'
                    context.set_details(msg)
                    context.set_code(grpc.StatusCode.RESOURCE_EXHAUSTED)
                    err = hangar_service_pb2.ErrorProto(code=8, message=msg)
                    totalSize, records = 0, []
                    yield hangar_service_pb2.FetchDataReply(error=err, raw_data=b'')
                    raise StopIteration()
        except StopIteration:
            totalSize, records = 0, []
        finally:
            # finish sending all remaining tensors if max size hash not been hit.
            if totalSize > 0:
                pack = chunks.serialize_record_pack(records)
                err = hangar_service_pb2.ErrorProto(code=0, message='OK')
                cIter = chunks.tensorChunkedIterator(
                    buf=pack, uncomp_nbytes=len(pack), itemsize=tensor.itemsize,
                    pb2_request=hangar_service_pb2.FetchDataReply, err=err)
                yield from cIter
            self.txnregister.abort_reader_txn(self.env.hashenv)

    def PushData(self, request_iterator, context):
        """Receive compressed streams of binary data from the client.

        In order to prevent errors or malicious behavior, the cryptographic hash
        of every tensor is calculated and compared to what the client "said" it
        is. If an error is detected, no sample in the entire stream will be
        saved to disk.
        """
        for idx, request in enumerate(request_iterator):
            if idx == 0:
                uncomp_nbytes = request.uncomp_nbytes
                comp_nbytes = request.comp_nbytes
                dBytes, offset = bytearray(comp_nbytes), 0
            size = len(request.raw_data)
            dBytes[offset: offset + size] = request.raw_data
            offset += size

        uncompBytes = blosc.decompress(dBytes)
        if uncomp_nbytes != len(uncompBytes):
            msg = f'ERROR: uncomp_nbytes sent: {uncomp_nbytes} != received {comp_nbytes}'
            context.set_details(msg)
            context.set_code(grpc.StatusCode.DATA_LOSS)
            err = hangar_service_pb2.ErrorProto(code=15, message=msg)
            reply = hangar_service_pb2.PushDataReply(error=err)
            return reply

        unpacked_records = chunks.deserialize_record_pack(uncompBytes)
        received_data = []
        for record in unpacked_records:
            data = chunks.deserialize_record(record)
            schema_hash = data.schema
            received_hash = ndarray_hasher_tcode_0(data.array)
            if received_hash != data.digest:
                msg = f'HASH MANGLED, received: {received_hash} != expected digest: {data.digest}'
                context.set_details(msg)
                context.set_code(grpc.StatusCode.DATA_LOSS)
                err = hangar_service_pb2.ErrorProto(code=15, message=msg)
                reply = hangar_service_pb2.PushDataReply(error=err)
                return reply
            received_data.append((received_hash, data.array))
        _ = self.CW.data(schema_hash, received_data)  # returns saved)_digests
        err = hangar_service_pb2.ErrorProto(code=0, message='OK')
        reply = hangar_service_pb2.PushDataReply(error=err)
        return reply

    # ------------------------ Fetch Find Missing -----------------------------------

    def FetchFindMissingCommits(self, request, context):
        """Determine commit digests existing on the server which are not present on the client.
        """
        c_branch_name = request.branch.name
        c_ordered_commits = request.commits

        try:
            s_history = summarize.list_history(
                refenv=self.env.refenv,
                branchenv=self.env.branchenv,
                branch_name=c_branch_name)
        except ValueError:
            msg = f'BRANCH NOT EXIST. Name: {c_branch_name}'
            context.set_code(grpc.StatusCode.NOT_FOUND)
            context.set_details(msg)
            err = hangar_service_pb2.ErrorProto(code=5, message=msg)
            reply = hangar_service_pb2.FindMissingCommitsReply(error=err)
            return reply

        s_orderset = set(s_history['order'])
        c_orderset = set(c_ordered_commits)
        c_missing = list(s_orderset.difference(c_orderset))   # only difference to PushFindMissingCommits

        err = hangar_service_pb2.ErrorProto(code=0, message='OK')
        if len(c_missing) == 0:
            brch = hangar_service_pb2.BranchRecord(name=c_branch_name, commit=s_history['head'])
            reply = hangar_service_pb2.FindMissingCommitsReply(branch=brch, error=err)
        else:
            brch = hangar_service_pb2.BranchRecord(name=c_branch_name, commit=s_history['head'])
            reply = hangar_service_pb2.FindMissingCommitsReply(branch=brch, error=err)
            reply.commits.extend(c_missing)

        return reply

    def PushFindMissingCommits(self, request, context):
        """Determine commit digests existing on the client which are not present on the server.
        """
        c_branch_name = request.branch.name
        c_head_commit = request.branch.commit
        c_ordered_commits = request.commits

        s_commits = commiting.list_all_commits(self.env.refenv)
        s_orderset = set(s_commits)
        c_orderset = set(c_ordered_commits)
        s_missing = list(c_orderset.difference(s_orderset))  # only difference to FetchFindMissingCommits

        err = hangar_service_pb2.ErrorProto(code=0, message='OK')
        if len(s_missing) == 0:
            brch = hangar_service_pb2.BranchRecord(name=c_branch_name, commit=c_head_commit)
            reply = hangar_service_pb2.FindMissingCommitsReply(branch=brch, error=err)
        else:
            brch = hangar_service_pb2.BranchRecord(name=c_branch_name, commit=c_head_commit)
            reply = hangar_service_pb2.FindMissingCommitsReply(branch=brch, error=err)
            reply.commits.extend(s_missing)

        return reply

    def FetchFindMissingHashRecords(self, request_iterator, context):
        """Determine data tensor hash records existing on the server and not on the client.
        """
        for idx, request in enumerate(request_iterator):
            if idx == 0:
                commit = request.commit
                hBytes, offset = bytearray(request.total_byte_size), 0
            size = len(request.hashs)
            hBytes[offset: offset + size] = request.hashs
            offset += size

        uncompBytes = blosc.decompress(hBytes)
        c_hashs_raw = chunks.deserialize_record_pack(uncompBytes)
        c_hashset = set([chunks.deserialize_ident(raw).digest for raw in c_hashs_raw])

        with tempfile.TemporaryDirectory() as tempD:
            tmpDF = os.path.join(tempD, 'test.lmdb')
            tmpDB = lmdb.open(path=tmpDF, **c.LMDB_SETTINGS)
            commiting.unpack_commit_ref(self.env.refenv, tmpDB, commit)
            s_hashes_schemas = queries.RecordQuery(tmpDB).data_hash_to_schema_hash()
            s_hashes = set(s_hashes_schemas.keys())
            tmpDB.close()

        c_missing = list(s_hashes.difference(c_hashset))
        c_hash_schemas_raw = [chunks.serialize_ident(c_mis, s_hashes_schemas[c_mis]) for c_mis in c_missing]
        raw_pack = chunks.serialize_record_pack(c_hash_schemas_raw)
        err = hangar_service_pb2.ErrorProto(code=0, message='OK')
        response_pb = hangar_service_pb2.FindMissingHashRecordsReply
        cIter = chunks.missingHashIterator(commit, raw_pack, err, response_pb)
        yield from cIter

    def PushFindMissingHashRecords(self, request_iterator, context):
        """Determine data tensor hash records existing on the client and not on the server.
        """
        for idx, request in enumerate(request_iterator):
            if idx == 0:
                commit = request.commit
                hBytes, offset = bytearray(request.total_byte_size), 0
            size = len(request.hashs)
            hBytes[offset: offset + size] = request.hashs
            offset += size

        uncompBytes = blosc.decompress(hBytes)
        c_hashs_raw = chunks.deserialize_record_pack(uncompBytes)
        c_hashset = set([chunks.deserialize_ident(raw).digest for raw in c_hashs_raw])

        s_hashset = set(hashs.HashQuery(self.env.hashenv).list_all_hash_keys_raw())
        s_missing = c_hashset.difference(s_hashset)
        s_hashs_raw = [chunks.serialize_ident(s_mis, '') for s_mis in s_missing]
        raw_pack = chunks.serialize_record_pack(s_hashs_raw)

        err = hangar_service_pb2.ErrorProto(code=0, message='OK')
        response_pb = hangar_service_pb2.FindMissingHashRecordsReply
        cIter = chunks.missingHashIterator(commit, raw_pack, err, response_pb)
        yield from cIter

    def FetchFindMissingSchemas(self, request, context):
        """Determine schema hash digest records existing on the server and not on the client.
        """
        commit = request.commit
        c_schemas = set(request.schema_digests)

        with tempfile.TemporaryDirectory() as tempD:
            tmpDF = os.path.join(tempD, 'test.lmdb')
            tmpDB = lmdb.open(path=tmpDF, **c.LMDB_SETTINGS)
            commiting.unpack_commit_ref(self.env.refenv, tmpDB, commit)
            s_schemas = set(queries.RecordQuery(tmpDB).schema_hashes())
            tmpDB.close()

        c_missing = list(s_schemas.difference(c_schemas))
        err = hangar_service_pb2.ErrorProto(code=0, message='OK')
        reply = hangar_service_pb2.FindMissingSchemasReply(commit=commit, error=err)
        reply.schema_digests.extend(c_missing)
        return reply

    def PushFindMissingSchemas(self, request, context):
        """Determine schema hash digest records existing on the client and not on the server.
        """
        commit = request.commit
        c_schemas = set(request.schema_digests)
        s_schemas = set(hashs.HashQuery(self.env.hashenv).list_all_schema_digests())
        s_missing = list(c_schemas.difference(s_schemas))

        err = hangar_service_pb2.ErrorProto(code=0, message='OK')
        reply = hangar_service_pb2.FindMissingSchemasReply(commit=commit, error=err)
        reply.schema_digests.extend(s_missing)
        return reply


def serve(hangar_path: str,
          overwrite: bool = False,
          *,
          channel_address: str = None,
          restrict_push: bool = None,
          username: str = None,
          password: str = None) -> tuple:
    """Start serving the GRPC server. Should only be called once.

    Raises:
        e: critical error from one of the workers.
    """

    # ------------------- Configure Server ------------------------------------

    dst_path = pjoin(hangar_path, c.DIR_HANGAR_SERVER, c.CONFIG_SERVER_NAME)
    CFG = configparser.ConfigParser()
    if os.path.isfile(dst_path):
        CFG.read(dst_path)
    else:
        src_path = pjoin(os.path.dirname(__file__), c.CONFIG_SERVER_NAME)
        CFG.read(src_path)

    serverCFG = CFG['SERVER_GRPC']
    enable_compression = serverCFG['enable_compression']
    if enable_compression == 'NoCompression':
        compression_val = grpc.Compression.NoCompression
    elif enable_compression == 'Deflate':
        compression_val = grpc.Compression.Deflate
    elif enable_compression == 'Gzip':
        compression_val = grpc.Compression.Gzip
    else:
        compression_val = grpc.Compression.NoCompression

    optimization_target = serverCFG['optimization_target']
    if channel_address is None:
        channel_address = serverCFG['channel_address']
    max_thread_pool_workers = int(serverCFG['max_thread_pool_workers'])
    max_concurrent_rpcs = int(serverCFG['max_concurrent_rpcs'])

    adminCFG = CFG['SERVER_ADMIN']
    if (restrict_push is None) and (username is None) and (password is None):
        admin_restrict_push = bool(int(adminCFG['restrict_push']))
        admin_username = adminCFG['username']
        admin_password = adminCFG['password']
    else:
        admin_restrict_push = restrict_push
        admin_username = username
        admin_password = password
    msg = 'PERMISSION ERROR: PUSH OPERATIONS RESTRICTED FOR CALLER'
    code = grpc.StatusCode.PERMISSION_DENIED
    interc = request_header_validator_interceptor.RequestHeaderValidatorInterceptor(
        admin_restrict_push, admin_username, admin_password, code, msg)

    # ---------------- Start the thread pool for the grpc server --------------

    grpc_thread_pool = futures.ThreadPoolExecutor(
        max_workers=max_thread_pool_workers,
        thread_name_prefix='grpc_thread_pool')
    server = grpc.server(
        thread_pool=grpc_thread_pool,
        maximum_concurrent_rpcs=max_concurrent_rpcs,
        options=[('grpc.optimization_target', optimization_target)],
        compression=compression_val,
        interceptors=(interc,))

    # ------------------- Start the GRPC server -------------------------------

    server_dir = pjoin(hangar_path, c.DIR_HANGAR_SERVER)
    hangserv = HangarServer(server_dir, overwrite)
    hangar_service_pb2_grpc.add_HangarServiceServicer_to_server(hangserv, server)
    server.add_insecure_port(channel_address)
    return (server, hangserv, channel_address)


if __name__ == '__main__':
    workdir = os.getcwd()
    print(workdir)
    serve(workdir)



import pytest
import numpy as np
from conftest import fixed_shape_backend_params


@pytest.mark.parametrize('backend', fixed_shape_backend_params)
def test_backend_property_reports_correct_backend(repo, array5by7, backend):

    wco = repo.checkout(write=True)
    aset = wco.add_ndarray_column('aset', prototype=array5by7, backend=backend)
    assert aset.backend == backend
    aset[0] = array5by7
    wco.commit('first')
    wco.close()

    rco = repo.checkout()
    naset = rco.columns['aset']
    assert naset.backend == backend
    rco.close()


@pytest.mark.parametrize('backend', fixed_shape_backend_params)
def test_setting_backend_property_cannot_change_backend(repo, array5by7, backend):

    wco = repo.checkout(write=True)
    aset = wco.add_ndarray_column('aset', prototype=array5by7, backend=backend)
    assert aset.backend == backend
    aset[0] = array5by7
    with pytest.raises(AttributeError):
        aset.backend = 'foo'
    wco.commit('first')
    wco.close()

    rco = repo.checkout()
    naset = rco.columns['aset']
    assert naset.backend == backend
    with pytest.raises(AttributeError):
        naset.backend = 'foo'
    rco.close()


@pytest.mark.parametrize('subsamples', [True, False])
@pytest.mark.parametrize('backend', fixed_shape_backend_params)
def test_setting_backend_opts_property_cannot_change_backend_opts(repo, array5by7, backend, subsamples):

    wco = repo.checkout(write=True)
    aset = wco.add_ndarray_column(
        'aset', prototype=array5by7, backend=backend, contains_subsamples=subsamples)
    if subsamples:
        aset.update({0: {0: array5by7}})
    else:
        aset[0] = array5by7
    with pytest.raises(AttributeError):
        aset.backend_options = {'foo': 'bar'}
    wco.commit('first')
    wco.close()

    rco = repo.checkout()
    naset = rco.columns['aset']
    assert naset.backend == backend
    with pytest.raises(AttributeError):
        naset.backend = {'foo': 'bar'}
    rco.close()


@pytest.mark.parametrize('shape,dtype,variable_shape,expected_backend', [
    [(10,), np.uint16, True, '10'],
    [(1000,), np.uint16, True, '00'],
    [(1000,), np.uint16, False, '00'],
    [(9_999_999,), np.uint8, False, '00'],
    [(10_000_000,), np.uint8, False, '00'],
    [(10_000_001,), np.uint8, False, '01'],
    [(10_000_001,), np.uint8, True, '00'],
    [(2, 2), np.uint16, True, '00'],
    [(2, 2), np.uint16, False, '01'],
    [(5, 2), np.uint16, True, '00'],
    [(5, 2), np.uint16, False, '01'],
])
@pytest.mark.parametrize('subsamples', [True, False])
def test_heuristics_select_backend(repo, shape, dtype, variable_shape, expected_backend, subsamples):
    wco = repo.checkout(write=True)
    prototype = np.ones(shape, dtype=dtype)
    aset = wco.add_ndarray_column(
        'aset', prototype=prototype, variable_shape=variable_shape, contains_subsamples=subsamples)
    assert aset.backend == expected_backend
    if subsamples:
        aset.update({'0': {'0': prototype}})
    else:
        aset['0'] = prototype
    wco.commit('first commit')
    assert aset.backend == expected_backend
    if subsamples:
        assert np.allclose(prototype, aset['0']['0'])
    else:
        assert np.allclose(prototype, aset['0'])
    wco.close()

    nwco = repo.checkout(write=True)
    naset = nwco.columns['aset']
    assert naset.backend == expected_backend
    if subsamples:
        assert np.allclose(prototype, naset['0']['0'])
    else:
        assert np.allclose(prototype, naset['0'])
    nwco.close()


@pytest.mark.parametrize('prototype', [np.random.randn(10), np.random.randn(1000), np.random.randn(2, 2)])
@pytest.mark.parametrize('backend', fixed_shape_backend_params)
@pytest.mark.parametrize('subsamples', [True, False])
def test_manual_override_heuristics_select_backend(repo, prototype, backend, subsamples):

    wco = repo.checkout(write=True)
    aset = wco.add_ndarray_column(
        'aset', prototype=prototype, backend=backend, contains_subsamples=subsamples)
    assert aset.backend == backend
    if subsamples:
        aset.update({'0': {'0': prototype}})
    else:
        aset['0'] = prototype
    wco.commit('first commit')
    assert aset.backend == backend
    if subsamples:
        assert np.allclose(prototype, aset['0']['0'])
    else:
        assert np.allclose(prototype, aset['0'])
    wco.close()

    nwco = repo.checkout(write=True)
    naset = nwco.columns['aset']
    assert naset.backend == backend
    if subsamples:
        assert np.allclose(prototype, naset['0']['0'])
    else:
        assert np.allclose(prototype, naset['0'])
    nwco.close()


def test_manual_override_heuristics_invalid_value_raises_error(repo):

    wco = repo.checkout(write=True)
    with pytest.raises(ValueError):
        wco.add_ndarray_column('aset', prototype=np.arange(10), backend='ERROR')
    wco.close()


@pytest.mark.parametrize('backendStart', fixed_shape_backend_params)
@pytest.mark.parametrize('backendEnd', fixed_shape_backend_params)
@pytest.mark.parametrize('subsamples', [True, False])
def test_manual_change_backends_after_write_works(repo, array5by7, backendStart, backendEnd, subsamples):

    wco = repo.checkout(write=True)
    aset = wco.add_ndarray_column(
        'aset', prototype=array5by7, backend=backendStart, contains_subsamples=subsamples)
    assert aset.backend == backendStart
    if subsamples:
        aset.update({0: {0: array5by7}})
    else:
        aset[0] = array5by7
    wco.commit('first commit')
    assert aset.backend == backendStart
    if subsamples:
        assert np.allclose(array5by7, aset[0][0])
    else:
        assert np.allclose(array5by7, aset[0])
    wco.close()

    nwco = repo.checkout(write=True)
    naset = nwco.columns['aset']
    assert naset.backend == backendStart

    naset.change_backend(backend=backendEnd)
    if subsamples:
        naset.update({1: {1: array5by7+1}})
    else:
        naset[1] = array5by7 + 1

    assert naset.backend == backendEnd
    if subsamples:
        assert np.allclose(array5by7, naset[0][0])
        assert np.allclose(array5by7+1, naset[1][1])
    else:
        assert np.allclose(array5by7, naset[0])
        assert np.allclose(array5by7+1, naset[1])
    nwco.commit('second')
    nwco.close()

    rco = repo.checkout()
    assert rco.columns['aset'].backend == backendEnd
    rco.close()


@pytest.mark.parametrize('backendStart', fixed_shape_backend_params)
@pytest.mark.parametrize('backendFail', ['lmao', '000'])
@pytest.mark.parametrize('subsamples', [True, False])
def test_manual_change_backend_to_invalid_fmt_code_fails(repo, array5by7, backendStart, backendFail, subsamples):

    wco = repo.checkout(write=True)
    aset = wco.add_ndarray_column(
        'aset', prototype=array5by7, backend=backendStart, contains_subsamples=subsamples)
    assert aset.backend == backendStart
    if subsamples:
        aset[0] = {0: array5by7}
    else:
        aset[0] = array5by7
    wco.commit('first commit')
    assert aset.backend == backendStart
    if subsamples:
        assert np.allclose(array5by7, aset[0][0])
    else:
        assert np.allclose(array5by7, aset[0])
    wco.close()

    nwco = repo.checkout(write=True)
    naset = nwco.columns['aset']
    assert naset.backend == backendStart

    with pytest.raises(ValueError):
        naset.change_backend(backend=backendFail)
    assert naset.backend == backendStart
    if subsamples:
        naset[1] = {1: array5by7+1}
    else:
        naset[1] = array5by7 + 1

    if subsamples:
        assert np.allclose(array5by7, naset[0][0])
        assert np.allclose(array5by7 + 1, naset[1][1])
    else:
        assert np.allclose(array5by7, naset[0])
        assert np.allclose(array5by7 + 1, naset[1])
    nwco.commit('second')
    nwco.close()


@pytest.mark.parametrize('backendStart', fixed_shape_backend_params)
@pytest.mark.parametrize('backendEnd', fixed_shape_backend_params)
@pytest.mark.parametrize('subsamples', [True, False])
def test_manual_change_backend_fails_while_in_cm(repo, array5by7, backendStart, backendEnd, subsamples):

    wco = repo.checkout(write=True)
    aset = wco.add_ndarray_column(
        'aset', prototype=array5by7, backend=backendStart, contains_subsamples=subsamples)
    assert aset.backend == backendStart
    if subsamples:
        aset[0] = {0: array5by7}
    else:
        aset[0] = array5by7
    wco.commit('first commit')
    assert aset.backend == backendStart
    if subsamples:
        assert np.allclose(array5by7, aset[0][0])
    else:
        assert np.allclose(array5by7, aset[0])
    wco.close()

    nwco = repo.checkout(write=True)
    naset = nwco.columns['aset']
    assert naset.backend == backendStart

    with nwco as c:
        with pytest.raises(RuntimeError):
            c['aset'].change_backend(backend=backendEnd)
        with pytest.raises(RuntimeError):
            naset.change_backend(backend=backendEnd)
        with pytest.raises(RuntimeError):
            c.columns['aset'].change_backend(backend=backendEnd)
        with pytest.raises(RuntimeError):
            nwco.columns['aset'].change_backend(backend=backendEnd)

    with naset as na:
        with pytest.raises(RuntimeError):
            na.change_backend(backend=backendEnd)
        with pytest.raises(RuntimeError):
            naset.change_backend(backend=backendEnd)
        with pytest.raises(RuntimeError):
            nwco.columns['aset'].change_backend(backend=backendEnd)

    assert naset.backend == backendStart
    if subsamples:
        naset[1] = {1: array5by7+1}
    else:
        naset[1] = array5by7 + 1

    if subsamples:
        assert np.allclose(array5by7, naset[0][0])
        assert np.allclose(array5by7 + 1, naset[1][1])
    else:
        assert np.allclose(array5by7, naset[0])
        assert np.allclose(array5by7 + 1, naset[1])
    nwco.commit('second')
    nwco.close()



@pytest.fixture(scope='class')
def dummy_writer_checkout(classrepo):
    wco = classrepo.checkout(write=True)
    yield wco
    wco.close()


class TestComplibRestrictions:

    @pytest.mark.parametrize('backend', ['01', '00'])
    @pytest.mark.parametrize('subsamples', [True, False])
    @pytest.mark.parametrize('complib', [
        'blosc:blosclz', 'blosc:lz4', 'blosc:lz4hc', 'blosc:zlib', 'blosc:zstd'
    ])
    @pytest.mark.parametrize('dtype,shape', [
        [np.float32, (1, 1, 1)],
        [np.float32, (3,)],
        [np.float64, (1,)],
        [np.uint8, (15,)],
        [np.uint8, (3, 2, 2)],
    ])
    def test_schema_smaller_16_bytes_cannot_select_blosc_backend(
        self, dummy_writer_checkout, backend, complib, dtype, shape, subsamples
    ):
        wco = dummy_writer_checkout
        be_opts = {'complib': complib, 'complevel': 3, 'shuffle': 'byte'}

        # prototype spec
        with pytest.raises(ValueError, match='blosc clib requires'):
            proto = np.zeros(shape, dtype=dtype)
            wco.add_ndarray_column(
                'aset', prototype=proto, backend=backend,
                backend_options=be_opts, contains_subsamples=subsamples)

        # shape and dtype spec
        with pytest.raises(ValueError, match='blosc clib requires'):
            wco.add_ndarray_column(
                'aset', shape=shape, dtype=dtype, backend=backend,
                backend_options=be_opts, contains_subsamples=subsamples)


@pytest.mark.parametrize('backend', ['01', '00'])
@pytest.mark.parametrize('subsamples', [True, False])
@pytest.mark.parametrize('dtype,shape', [
    [np.float32, (1, 1, 1)],
    [np.float32, (3,)],
    [np.float64, (1,)],
    [np.uint8, (15,)],
    [np.uint8, (3, 2, 2)],
])
def test_schema_smaller_16_bytes_does_not_use_heuristic_to_select_blosc(
    repo, backend, dtype, shape, subsamples
):
    wco = repo.checkout(write=True)
    proto = np.zeros(shape, dtype=dtype)
    aset = wco.add_ndarray_column(
        'aset', prototype=proto, backend=backend, contains_subsamples=subsamples)
    bad_clibs = ['blosc:blosclz', 'blosc:lz4', 'blosc:lz4hc', 'blosc:zlib', 'blosc:zstd']
    assert aset.backend_options['complib'] not in bad_clibs
    if subsamples:
        aset[0] = {0: proto}
    else:
        aset[0] = proto
    assert aset.backend_options['complib'] not in bad_clibs
    wco.close()


@pytest.mark.parametrize('backend', ['01', '00'])
@pytest.mark.parametrize('subsamples', [True, False])
@pytest.mark.parametrize('complib', [
    'blosc:blosclz', 'blosc:lz4', 'blosc:lz4hc', 'blosc:zlib', 'blosc:zstd'
])
@pytest.mark.parametrize('dtype,shape', [
    [np.float32, (1, 1, 1)],
    [np.float32, (3,)],
    [np.float64, (1,)],
    [np.uint8, (15,)],
    [np.uint8, (3, 2, 2)],
])
def test_schema_smaller_16_bytes_cannot_change_to_blosc_backend(
    repo, backend, complib, shape, dtype, subsamples):

    wco = repo.checkout(write=True)
    aset = wco.add_ndarray_column(
        'aset', shape=shape, dtype=dtype, backend=backend, contains_subsamples=subsamples)
    proto = np.zeros(shape, dtype=dtype)
    if subsamples:
        aset[0] = {0: proto}
    else:
        aset[0] = proto

    be_opts = {'complib': complib, 'complevel': 3, 'shuffle': None}
    with pytest.raises(ValueError, match='blosc clib requires'):
        aset.change_backend(backend=backend, backend_options=be_opts)
    wco.close()



import array
from cpython cimport array

import numpy as np
from hashlib import blake2b


cpdef str hash_type_code_from_digest(str digest):
    return digest[0]


cpdef object hash_func_from_tcode(str tcode):
    if tcode == '0':
        return ndarray_hasher_tcode_0
    elif tcode == '1':
        return schema_hasher_tcode_1
    elif tcode == '2':
        return pystr_hasher_tcode_2
    else:
        raise ValueError(f'unknown hash function type code. tcode: {tcode}')


# ---------------------------- numpy ndarray data ------------------------------


cdef bytes ser_int_list(list lst):
    cdef Py_ssize_t n=len(lst)
    cdef array.array res=array.array('i')

    array.resize(res, n)  #preallocate memory
    for i in range(n):
        # lst.__get__() needs Python-Integer, so let i
        # be a python-integer (not cdef)
        res.data.as_ints[i] = lst[i]
    return res.data.as_chars[:n*sizeof(int)]


def ndarray_hasher_tcode_0(array not None):
    """Generate the hex digest of some array data.

    This method hashes the concatenation of both array data bytes as well as a
    binary struct with the array shape and dtype num packed in. This is in
    order to avoid hash collisions where an array can have the same bytes, but
    different shape. an example of a collision is: np.zeros((10, 10, 10)) and
    np.zeros((1000,))

    Parameters
    ----------
    array : np.ndarray
        array data to take the hash of

    Returns
    -------
    str
        hex digest of the array data with typecode prepended by '{tcode}='.
    """
    cdef str digest
    cdef bytes other_info
    cdef list shape = []

    shape = list(array.shape)
    shape.append(array.dtype.num)
    other_info = ser_int_list(shape)

    hasher = blake2b(array, digest_size=20)
    hasher.update(other_info)
    digest = hasher.hexdigest()
    return f'0={digest}'


# ------------------------------ Schema ---------------------------------------


def _make_hashable(o):
    """Sort container object and deterministically output frozen representation
    """
    if isinstance(o, (tuple, list)):
        return tuple((_make_hashable(e) for e in o))

    if isinstance(o, dict):
        return tuple(sorted((k, _make_hashable(v)) for k, v in o.items()))

    if isinstance(o, (set, frozenset)):
        return tuple(sorted(_make_hashable(e) for e in o))
    return o


cpdef str schema_hasher_tcode_1(dict schema):
    """Generate the schema hash for some schema specification

    Parameters
    ----------
    schema : dict
        dict representation of the schema spec.

    Returns
    -------
    str
        hex digest of this information with typecode prepended by '{tcode}='.
    """
    cdef bytes serialized
    cdef str digest, res

    frozenschema = _make_hashable(schema)
    serialized = repr(frozenschema).encode()
    digest = blake2b(serialized, digest_size=6).hexdigest()
    res = f'1={digest}'
    return res


# --------------------------- string type data ----------------------------------------


cpdef str pystr_hasher_tcode_2(str value):
    """Generate the hash digest of some metadata value

    Parameters
    ----------
    value : str
        data to set as the value of the metadata key.

    Returns
    -------
    str
        hex digest of the metadata value with typecode prepended by '{tcode}='.
    """
    cdef bytes raw
    cdef str digest, res

    raw = value.encode()
    digest = blake2b(raw, digest_size=20).hexdigest()
    res = f'2={digest}'
    return res



import pytest
import numpy as np
from conftest import fixed_shape_backend_params, variable_shape_backend_params
from itertools import permutations


def assert_equal(arr, arr2):
    assert np.array_equal(arr, arr2)
    assert arr.dtype == arr2.dtype


class TestColumn(object):

    @pytest.mark.parametrize('name', [
        'invalid\n', '\ninvalid', 'inv name', 'inva@lid', 12, ' try', 'andthis ',
        'VeryLongNameIsInvalidOver64CharactersNotAllowedVeryLongNameIsInva'])
    def test_invalid_column_name(self, repo, randomsizedarray, name):
        co = repo.checkout(write=True)
        with pytest.raises(ValueError):
            co.add_ndarray_column(name=name, prototype=randomsizedarray)
        with pytest.raises(ValueError):
            co.add_str_column(name=name)
        co.close()

    def test_read_only_mode(self, aset_samples_initialized_repo):
        import hangar
        co = aset_samples_initialized_repo.checkout()
        assert isinstance(co, hangar.checkout.ReaderCheckout)
        with pytest.raises(AttributeError):
            assert co.add_ndarray_column('foo')
        with pytest.raises(AttributeError):
            assert co.add_str_column('foo')
        with pytest.raises(PermissionError):
            del co.columns['foo']
        with pytest.raises(PermissionError):
            del co.columns['foo']
        assert len(co.columns['writtenaset']) == 0
        co.close()

    def test_get_column(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)

        # getting the column with `get`
        asetOld = co.columns.get('writtenaset')
        asetOldPath = asetOld._path
        asetOldAsetn = asetOld.column
        asetOldDefaultSchemaHash = asetOld._schema.schema_hash_digest()

        asetOld['1'] = array5by7
        co.commit('this is a commit message')
        co.close()
        co = aset_samples_initialized_repo.checkout()

        # getting column with dictionary like style method
        asetNew = co.columns['writtenaset']
        assert_equal(asetNew['1'], array5by7)
        assert asetOldPath == asetNew._path
        assert asetOldAsetn == asetNew.column
        assert asetOldDefaultSchemaHash == asetNew._schema.schema_hash_digest()
        co.close()

    @pytest.mark.parametrize("aset_backend", fixed_shape_backend_params)
    def test_remove_column(self, aset_backend, aset_samples_initialized_repo):
        co = aset_samples_initialized_repo.checkout(write=True)
        del co.columns['writtenaset']
        with pytest.raises(KeyError):
            del co.columns['writtenaset']

        co.add_ndarray_column('writtenaset', shape=(5, 7), dtype=np.float64, backend=aset_backend)
        assert len(co.columns) == 1
        del co.columns['writtenaset']
        co.commit('this is a commit message')
        co.close()

        co = aset_samples_initialized_repo.checkout(write=True)
        assert len(co.columns) == 0

        co.add_ndarray_column('writtenaset', shape=(5, 7), dtype=np.float64, backend=aset_backend)
        co.commit('this is a commit message')
        co.close()
        co = aset_samples_initialized_repo.checkout(write=True)
        assert len(co.columns) == 1
        del co.columns['writtenaset']
        assert len(co.columns) == 0
        co.commit('this is a commit message')
        co.close()

    @pytest.mark.parametrize("aset_backend", fixed_shape_backend_params)
    def test_init_again(self, aset_backend, repo, randomsizedarray):
        co = repo.checkout(write=True)
        co.add_ndarray_column('aset', prototype=randomsizedarray, backend=aset_backend)
        with pytest.raises(LookupError):
            co.add_ndarray_column('aset', prototype=randomsizedarray, backend=aset_backend)
        co.close()

    @pytest.mark.parametrize("aset_backend", fixed_shape_backend_params)
    def test_column_with_more_dimension(self, aset_backend, repo):
        co = repo.checkout(write=True)
        shape = (0, 1, 2)
        with pytest.raises(ValueError):
            co.add_ndarray_column('aset', shape=shape, dtype=np.int, backend=aset_backend)
        shape = [1] * 31
        aset = co.add_ndarray_column('aset1', shape=shape, dtype=np.int, backend=aset_backend)
        assert len(aset.shape) == 31
        shape = [1] * 32
        with pytest.raises(ValueError):
            # maximum tensor rank must be <= 31
            co.add_ndarray_column('aset2', shape=shape, dtype=np.int, backend=aset_backend)
        co.close()

    @pytest.mark.parametrize("aset_backend", fixed_shape_backend_params)
    def test_column_with_empty_dimension(self, aset_backend, repo):
        co = repo.checkout(write=True)
        arr = np.array(1, dtype=np.int64)
        aset = co.add_ndarray_column('aset1', shape=(), dtype=np.int64, backend=aset_backend)
        aset['1'] = arr
        co.commit('this is a commit message')
        aset = co.add_ndarray_column('aset2', prototype=arr)
        aset['1'] = arr
        co.commit('this is a commit message')
        co.close()
        co = repo.checkout()
        aset1 = co.columns['aset1']
        aset2 = co.columns['aset2']
        assert_equal(aset1['1'], arr)
        assert_equal(aset2['1'], arr)
        co.close()

    @pytest.mark.parametrize("aset_backend", fixed_shape_backend_params)
    def test_column_with_int_specifier_as_dimension(self, aset_backend, repo):
        co = repo.checkout(write=True)
        arr = np.arange(10, dtype=np.int64)
        aset = co.add_ndarray_column('aset1', shape=10, dtype=np.int64, backend=aset_backend)
        aset['1'] = arr
        co.commit('this is a commit message')
        arr2 = np.array(53, dtype=np.int64)
        aset = co.add_ndarray_column('aset2', prototype=arr2)
        aset['1'] = arr2
        co.commit('this is a commit message')
        co.close()
        co = repo.checkout()
        aset1 = co.columns['aset1']
        aset2 = co.columns['aset2']
        assert_equal(aset1['1'], arr)
        assert_equal(aset2['1'], arr2)
        co.close()

    @pytest.mark.parametrize('write', [True, False])
    @pytest.mark.parametrize("aset_backend", fixed_shape_backend_params)
    def test_getattr_does_not_raise_permission_error_if_alive(self, aset_backend, write, repo):
        co = repo.checkout(write=True)
        arr = np.arange(10, dtype=np.int64)
        aset = co.add_ndarray_column('aset1', shape=10, dtype=np.int64, backend=aset_backend)
        aset['1'] = arr
        co.commit('hello')
        co.close()
        co = repo.checkout(write=write)
        aset = co.columns['aset1']

        assert hasattr(aset, 'doesnotexist') is False  # does not raise error
        assert hasattr(aset, '_mode') is True
        with pytest.raises(AttributeError):
            assert getattr(aset, 'doesnotexist')
        assert getattr(aset, '_mode') == 'a' if write else 'r'

        co.close()
        with pytest.raises(PermissionError):
            hasattr(aset, 'doesnotexist')
        with pytest.raises(PermissionError):
            hasattr(aset, '_mode')


class TestDataWithFixedSizedColumn(object):

    @pytest.mark.parametrize("aset1_backend", fixed_shape_backend_params)
    @pytest.mark.parametrize("aset2_backend", fixed_shape_backend_params)
    @pytest.mark.parametrize("aset3_backend", fixed_shape_backend_params)
    def test_column_remote_references_property_with_none(
            self, aset1_backend, aset2_backend, aset3_backend, repo, randomsizedarray
    ):
        co = repo.checkout(write=True)
        aset1 = co.add_ndarray_column('aset1', prototype=randomsizedarray, backend=aset1_backend)
        aset2 = co.add_ndarray_column('aset2', shape=(2, 2), dtype=np.int, backend=aset2_backend)
        aset3 = co.add_ndarray_column('aset3', shape=(3, 4), dtype=np.float32, backend=aset3_backend)

        with aset1 as d1, aset2 as d2, aset3 as d3:
            d1[1] = randomsizedarray
            d2[1] = np.ones((2, 2), dtype=np.int)
            d3[1] = np.ones((3, 4), dtype=np.float32)

        assert co.columns.contains_remote_references == {'aset1': False, 'aset2': False, 'aset3': False}
        assert co.columns.remote_sample_keys == {'aset1': (), 'aset2': (), 'aset3': ()}
        co.close()

    @pytest.mark.parametrize("aset1_backend", fixed_shape_backend_params)
    @pytest.mark.parametrize("aset2_backend", fixed_shape_backend_params)
    @pytest.mark.parametrize("aset3_backend", fixed_shape_backend_params)
    def test_column_remote_references_property_with_remotes(
            self, aset1_backend, aset2_backend, aset3_backend, repo, randomsizedarray
    ):
        co = repo.checkout(write=True)
        aset1 = co.add_ndarray_column('aset1', prototype=randomsizedarray, backend=aset1_backend)
        aset2 = co.add_ndarray_column('aset2', shape=(2, 2), dtype=np.int, backend=aset2_backend)
        aset3 = co.add_ndarray_column('aset3', shape=(3, 4), dtype=np.float32, backend=aset3_backend)

        with aset1 as d1, aset2 as d2, aset3 as d3:
            d1[1] = randomsizedarray
            d2[1] = np.ones((2, 2), dtype=np.int)
            d3[1] = np.ones((3, 4), dtype=np.float32)

        assert co.columns.contains_remote_references == {'aset1': False, 'aset2': False, 'aset3': False}
        assert co.columns.remote_sample_keys == {'aset1': (), 'aset2': (), 'aset3': ()}
        co.commit('hello')
        co.close()
        co = repo.checkout()

        # perform the mock
        from hangar.backends import backend_decoder
        template = backend_decoder(b'50:daeaaeeaebv')
        co._columns._columns['aset1']._samples[12] = template
        co._columns._columns['aset2']._samples[22] = template

        assert co.columns.contains_remote_references == {'aset1': True, 'aset2': True, 'aset3': False}
        assert co.columns.remote_sample_keys == {'aset1': (12,), 'aset2': (22,), 'aset3': ()}
        co.close()

    @pytest.mark.parametrize("aset1_backend", fixed_shape_backend_params)
    @pytest.mark.parametrize("aset2_backend", fixed_shape_backend_params)
    @pytest.mark.parametrize("aset3_backend", fixed_shape_backend_params)
    def test_iterating_over(self, aset1_backend, aset2_backend, aset3_backend, repo, randomsizedarray):
        co = repo.checkout(write=True)
        all_tensors = []
        aset1 = co.add_ndarray_column('aset1', prototype=randomsizedarray, backend=aset1_backend)
        aset2 = co.add_ndarray_column('aset2', shape=(2, 2), dtype=np.int, backend=aset2_backend)
        aset3 = co.add_ndarray_column('aset3', shape=(3, 4), dtype=np.float32, backend=aset3_backend)

        with aset1 as d1, aset2 as d2, aset3 as d3:
            d1['1'] = randomsizedarray
            d1['2'] = np.zeros_like(randomsizedarray)
            d1['3'] = np.zeros_like(randomsizedarray) + 5

            d2['1'] = np.ones((2, 2), dtype=np.int)
            d2['2'] = np.ones((2, 2), dtype=np.int) * 5
            d2['3'] = np.zeros((2, 2), dtype=np.int)

            d3['1'] = np.ones((3, 4), dtype=np.float32)
            d3['2'] = np.ones((3, 4), dtype=np.float32) * 7
            d3['3'] = np.zeros((3, 4), dtype=np.float32)

        all_tensors.extend([aset1['1'], aset1['2'], aset1['3']])
        all_tensors.extend([aset2['1'], aset2['2'], aset2['3']])
        all_tensors.extend([aset3['1'], aset3['2'], aset3['3']])

        co.commit('this is a commit message')
        co.close()
        co = repo.checkout()
        # iterating over .items()
        tensors_in_the_order = iter(all_tensors)
        for dname, aset in co.columns.items():
            assert aset._column_name == dname
            for sname, sample in aset.items():
                assert_equal(sample, next(tensors_in_the_order))

        # iterating over .keys()
        tensors_in_the_order = iter(all_tensors)
        for dname in co.columns.keys():
            for sname in co.columns[dname].keys():
                assert_equal(co.columns[dname][sname], next(tensors_in_the_order))

        # iterating over .values()
        tensors_in_the_order = iter(all_tensors)
        for aset in co.columns.values():
            for sample in aset.values():
                assert_equal(sample, next(tensors_in_the_order))
        co.close()

    @pytest.mark.parametrize("aset1_backend", fixed_shape_backend_params)
    @pytest.mark.parametrize("aset2_backend", fixed_shape_backend_params)
    @pytest.mark.parametrize("aset3_backend", fixed_shape_backend_params)
    def test_iterating_over_local_only(self, aset1_backend, aset2_backend, aset3_backend, repo, randomsizedarray):
        co = repo.checkout(write=True)
        all_tensors = []
        aset1 = co.add_ndarray_column('aset1', prototype=randomsizedarray, backend=aset1_backend)
        aset2 = co.add_ndarray_column('aset2', shape=(2, 2), dtype=np.int, backend=aset2_backend)
        aset3 = co.add_ndarray_column('aset3', shape=(3, 4), dtype=np.float32, backend=aset3_backend)

        with aset1 as d1, aset2 as d2, aset3 as d3:
            d1['1'] = randomsizedarray
            d1['2'] = np.zeros_like(randomsizedarray)
            d1['3'] = np.zeros_like(randomsizedarray) + 5

            d2['1'] = np.ones((2, 2), dtype=np.int)
            d2['2'] = np.ones((2, 2), dtype=np.int) * 5
            d2['3'] = np.zeros((2, 2), dtype=np.int)

            d3['1'] = np.ones((3, 4), dtype=np.float32)
            d3['2'] = np.ones((3, 4), dtype=np.float32) * 7
            d3['3'] = np.zeros((3, 4), dtype=np.float32)

        all_tensors.extend([aset1['1'], aset1['2'], aset1['3']])
        all_tensors.extend([aset2['1'], aset2['2'], aset2['3']])
        all_tensors.extend([aset3['1'], aset3['2'], aset3['3']])

        co.commit('this is a commit message')
        co.close()
        co = repo.checkout()

        # perform the mock
        from hangar.backends import backend_decoder
        template = backend_decoder(b'50:daeaaeeaebv')
        co._columns._columns['aset1']._samples['4'] = template
        co._columns._columns['aset2']._samples['4'] = template

        # iterating over .items()
        tensors_in_the_order = iter(all_tensors)
        for dname in ['aset1', 'aset2', 'aset3']:
            aset = co.columns[dname]
            count = 0
            for sname, sample in aset.items(local=True):
                count += 1
                assert_equal(sample, next(tensors_in_the_order))
                assert '4' != sname
            assert count == 3

        # iterating over .keys()
        tensors_in_the_order = iter(all_tensors)
        for dname in ['aset1', 'aset2', 'aset3']:
            aset = co.columns[dname]
            count = 0
            for sname in aset.keys(local=True):
                count += 1
                assert_equal(aset[sname], next(tensors_in_the_order))
                assert '4' != sname
            assert count == 3

        # iterating over .values()
        tensors_in_the_order = iter(all_tensors)
        for dname in ['aset1', 'aset2', 'aset3']:
            aset = co.columns[dname]
            count = 0
            for sample in aset.values(local=True):
                count += 1
                assert_equal(sample, next(tensors_in_the_order))
            assert count == 3

        assert list(co['aset1'].keys()) == ['1', '2', '3', '4']
        with pytest.raises((FileNotFoundError, KeyError)):
            list(co['aset1'].values())
        with pytest.raises((FileNotFoundError, KeyError)):
            list(co['aset1'].items())

        assert list(co['aset2'].keys()) == ['1', '2', '3', '4']
        with pytest.raises((FileNotFoundError, KeyError)):
            list(co['aset2'].values())
        with pytest.raises((FileNotFoundError, KeyError)):
            list(co['aset2'].items())

        assert list(co['aset3'].keys()) == ['1', '2', '3']
        assert len(list(co['aset3'].values())) == 3
        assert len(list(co['aset3'].items())) == 3
        co.close()

    def test_get_data(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        co.columns['writtenaset']['1'] = array5by7
        co.commit('this is a commit message')
        co.close()
        co = aset_samples_initialized_repo.checkout()
        assert np.allclose(co.columns['writtenaset']['1'], co.columns.get('writtenaset').get('1'), array5by7)
        co.close()

    def test_get_sample_with_default_works(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        res = co.columns['writtenaset'].get('doesnotexist', default=500)
        assert res is 500
        res = co.columns['writtenaset'].get('doesnotexist', 500)
        assert res is 500
        co.close()

    def test_get_multiple_samples_fails(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        co.columns['writtenaset']['1'] = array5by7
        co.columns['writtenaset']['2'] = array5by7 + 1
        co.columns['writtenaset']['3'] = array5by7 + 2
        co.commit('this is a commit message')
        co.close()

        nco = aset_samples_initialized_repo.checkout()
        with pytest.raises(TypeError):
            res = nco.columns['writtenaset'].get(['1', '2'])
        res = nco.columns['writtenaset'].get(('1', '2'))
        assert res is None

        aset = nco.columns['writtenaset']
        with pytest.raises(TypeError):
            res = aset.get(*('1', '2', '3'))
        nco.close()

    def test_getitem_multiple_samples_missing_key(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        co.columns['writtenaset']['1'] = array5by7
        co.commit('this is a commit message')
        co.close()

        nco = aset_samples_initialized_repo.checkout()
        with pytest.raises(KeyError):
            nco.columns['writtenaset'][('1', '2')]
        with pytest.raises(KeyError):
            aset = nco.columns['writtenaset']
            aset[('1', '2')]
        nco.close()

    def test_get_multiple_samples_missing_key(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        co.columns['writtenaset']['1'] = array5by7
        co.commit('this is a commit message')
        co.close()

        nco = aset_samples_initialized_repo.checkout()
        aset = nco.columns['writtenaset']
        res = aset.get(('1', '2'))
        assert res == None
        nco.close()

    def test_add_data_str_keys(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        aset = co.columns['writtenaset']
        with pytest.raises(KeyError):
            aset['somerandomkey']

        aset['1'] = array5by7
        aset['2'] = array5by7
        co.commit('this is a commit message')
        co.close()
        co = aset_samples_initialized_repo.checkout()
        assert_equal(co.columns['writtenaset']['1'], array5by7)
        assert_equal(co.columns['writtenaset']['2'], array5by7)
        co.close()

    def test_add_data_int_keys(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        aset = co.columns['writtenaset']

        aset[1] = array5by7
        secondArray = array5by7 + 1
        aset[2] = secondArray
        co.commit('this is a commit message')
        co.close()
        co = aset_samples_initialized_repo.checkout()
        assert_equal(co.columns['writtenaset'][1], array5by7)
        assert_equal(co.columns['writtenaset'][2], secondArray)
        co.close()

    def test_cannot_add_data_negative_int_key(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        aset = co.columns['writtenaset']
        with pytest.raises(ValueError):
            aset[-1] = array5by7
        assert len(co.columns['writtenaset']) == 0
        co.close()

    def test_cannot_add_data_float_key(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        aset = co.columns['writtenaset']
        with pytest.raises(ValueError):
            aset[2.1] = array5by7
        with pytest.raises(ValueError):
            aset[0.0] = array5by7
        assert len(co.columns['writtenaset']) == 0
        co.close()

    def test_add_data_mixed_int_str_keys(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        aset = co.columns['writtenaset']

        aset[1] = array5by7
        newFirstArray = array5by7 + 1
        aset['1'] = newFirstArray
        secondArray = array5by7 + 2
        aset[2] = secondArray
        thirdArray = array5by7 + 3
        aset['2'] = thirdArray
        co.commit('this is a commit message')
        co.close()
        co = aset_samples_initialized_repo.checkout()
        assert_equal(co.columns['writtenaset'][1], array5by7)
        assert_equal(co.columns['writtenaset']['1'], newFirstArray)
        assert_equal(co.columns['writtenaset'][2], secondArray)
        assert_equal(co.columns['writtenaset']['2'], thirdArray)
        co.close()

    def test_cannot_add_data_sample_name_longer_than_64_characters(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        aset = co.columns['writtenaset']
        with pytest.raises(ValueError):
            aset['VeryLongNameIsInvalidOver64CharactersNotAllowedVeryLongNameIsInva'] = array5by7
        assert len(co.columns['writtenaset']) == 0
        co.close()

    def test_add_with_wrong_argument_order(self, aset_samples_initialized_w_checkout, array5by7):
        aset = aset_samples_initialized_w_checkout.columns['writtenaset']
        with pytest.raises(ValueError):
            aset[array5by7] = '1'

    def test_update_with_dict_single_item(self, aset_samples_initialized_w_checkout, array5by7):
        aset = aset_samples_initialized_w_checkout.columns['writtenaset']
        data_map = {'foo': array5by7}
        aset.update(data_map)
        assert_equal(aset['foo'], array5by7)

    def test_update_with_dict_multiple_items(self, aset_samples_initialized_w_checkout, array5by7):
        aset = aset_samples_initialized_w_checkout.columns['writtenaset']
        data_map = {
            'foo': array5by7,
            1: array5by7+1
        }
        aset.update(data_map)
        assert_equal(aset['foo'], array5by7)
        assert_equal(aset[1], array5by7+1)

    def test_update_with_list_single_item(self, aset_samples_initialized_w_checkout, array5by7):
        aset = aset_samples_initialized_w_checkout.columns['writtenaset']
        data_map = ['foo', array5by7]
        with pytest.raises(ValueError, match='dictionary update sequence'):
            aset.update(data_map)
        assert 'foo' not in aset

        aset.update((data_map,))  # try again while contained in iterable
        assert_equal(aset['foo'], array5by7)

    def test_update_with_list_multiple_items(self, aset_samples_initialized_w_checkout, array5by7):
        aset = aset_samples_initialized_w_checkout.columns['writtenaset']
        data_map = [
            ('foo', array5by7),
            (1, array5by7+1),
        ]
        aset.update(data_map)
        assert_equal(aset['foo'], array5by7)
        assert_equal(aset[1], array5by7+1)

    def test_update_with_only_kwargs_single_item(self, aset_samples_initialized_w_checkout, array5by7):
        aset = aset_samples_initialized_w_checkout.columns['writtenaset']
        aset.update(foo=array5by7)
        assert_equal(aset['foo'], array5by7)

    def test_update_with_only_kwargs_multiple_items(self, aset_samples_initialized_w_checkout, array5by7):
        aset = aset_samples_initialized_w_checkout.columns['writtenaset']
        aset.update(foo=array5by7, bar=array5by7+1)
        assert_equal(aset['foo'], array5by7)
        assert_equal(aset['bar'], array5by7+1)

    def test_update_with_list_and_kwargs(self, aset_samples_initialized_w_checkout, array5by7):
        aset = aset_samples_initialized_w_checkout.columns['writtenaset']
        data_map = [
            ('foo', array5by7),
            (1, array5by7+1),
        ]
        aset.update(data_map, bar=array5by7+2)
        assert_equal(aset['foo'], array5by7)
        assert_equal(aset[1], array5by7+1)
        assert_equal(aset['bar'], array5by7 + 2)

    def test_update_with_dict_and_kwargs(self, aset_samples_initialized_w_checkout, array5by7):
        aset = aset_samples_initialized_w_checkout.columns['writtenaset']
        data_map = {
            'foo': array5by7,
            1: array5by7+1,
        }
        aset.update(data_map, bar=array5by7+2)
        assert_equal(aset['foo'], array5by7)
        assert_equal(aset[1], array5by7+1)
        assert_equal(aset['bar'], array5by7 + 2)

    def test_update_with_dict_and_kwargs_does_not_modify_input_in_calling_scopy(
        self, aset_samples_initialized_w_checkout, array5by7
    ):
        """ensure bug does not revert.

        Had a case where if dict was passed as ``other`` along with kwargs, the operation
        would complete as normally, but when control returned to the caller the original
        dict passed in as ``other`` would have been silently merged with the kwargs.
        """
        aset = aset_samples_initialized_w_checkout.columns['writtenaset']
        data_map = {
            'foo': array5by7,
            1: array5by7+1,
        }
        data_map_before = list(data_map.keys())
        aset.update(data_map, bar=array5by7+2)
        # in bug case, would now observe that data_map would have been
        # silently modified in a method analogous to calling:
        #
        #   ``data_map.update({'bar': np.array})``
        #
        assert list(data_map.keys()) == data_map_before

    @pytest.mark.parametrize('data_map', [
        ['foo', {'bar': np.random.random((5, 7))}],
        ['foo', 'bar', np.random.random((5, 7))],
        [('foo', 'bar', np.random.random((5, 7)))],
        [{('foo', 'bar'): np.random.random((5, 7))}],
        [('foo', 'bar', np.random.random((5, 7)))],
        [('foo', np.random.random((5, 7)), 'bar')],
        [(np.random.random((5, 7)), 'foo', 'bar')],
        [('foo', np.random.random((5, 7)), 'bar'), ('valid', np.random.random((5, 7)))],
        [('valid', np.random.random((5, 7))), ('foo', np.random.random((5, 7)), 'bar')],
        {'foo': np.random.random((5, 7)), 'bar': (np.random.random((5, 7)), np.random.random((5, 7)))},
    ])
    def test_update_with_invalid_data_map_fails(self, aset_samples_initialized_w_checkout, data_map):
        aset = aset_samples_initialized_w_checkout['writtenaset']
        with pytest.raises(ValueError):
            aset.update(data_map)

    @pytest.mark.parametrize('key,value', [
        ['foo', {'bar': np.random.random((5, 7))}],
        ['foo', ('bar', np.random.random((5, 7)))],
        [('foo', 'bar'), np.random.random((5, 7))],
        ['foo', {('foo', 'bar'): np.random.random((5, 7))}],
        ['foo', ('bar', np.random.random((5, 7)))],
        ['foo', (np.random.random((5, 7)), 'bar')],
        [np.random.random((5, 7)), ('foo', 'bar')],
        [('foo', np.random.random((5, 7)), 'bar'), ('valid', np.random.random((5, 7)))],
        [('valid', np.random.random((5, 7))), ('foo', np.random.random((5, 7)), 'bar')],
        [('valid', np.random.random((5, 7))), ('valid2', np.random.random((5, 7)))],
    ])
    def test_setitem_with_invalid_data_map_fails(self, aset_samples_initialized_w_checkout, key, value):
        aset = aset_samples_initialized_w_checkout['writtenaset']
        with pytest.raises(ValueError):
            aset[key] = value

    def test_add_multiple_data_single_commit(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        co.columns['writtenaset']['1'] = array5by7
        new_array = np.zeros_like(array5by7)
        co.columns['writtenaset']['2'] = new_array
        co.commit('this is a commit message')
        co.close()

        co = aset_samples_initialized_repo.checkout()
        aset = co.columns['writtenaset']
        assert len(aset) == 2
        assert list(aset.keys()) == ['1', '2']
        assert_equal(aset['1'], array5by7)
        co.close()

    def test_add_same_data_same_key_does_not_duplicate_hash(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        aset = co.columns['writtenaset']
        aset['1'] = array5by7
        old_spec = aset._samples['1']
        aset['1'] = array5by7
        new_spec = aset._samples['1']
        assert old_spec == new_spec
        assert len(aset) == 1
        assert len(aset._samples) == 1
        co.close()

    def test_multiple_data_multiple_commit(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        co.columns['writtenaset']['1'] = array5by7
        co.commit('this is a commit message')
        new_array = np.zeros_like(array5by7)
        co.columns['writtenaset']['2'] = new_array
        co.close()

        new_new_array = new_array + 5
        co = aset_samples_initialized_repo.checkout(write=True)
        co.columns['writtenaset']['3'] = new_new_array
        co.commit('this is a commit message')
        co.close()

        co = aset_samples_initialized_repo.checkout()
        aset = co.columns['writtenaset']
        assert_equal(aset['1'], array5by7)
        assert_equal(aset['2'], new_array)
        assert_equal(aset['3'], new_new_array)
        co.close()

    def test_added_but_not_commited(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        co.columns['writtenaset']['1'] = array5by7
        co.close()

        with pytest.raises(PermissionError):
            co.commit('this is a commit message')

        co = aset_samples_initialized_repo.checkout()
        aset = co.columns['writtenaset']
        with pytest.raises(KeyError):
            aset['1']
        co.close()

    def test_remove_data(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        co.columns['writtenaset']['1'] = array5by7
        co.columns['writtenaset']['2'] = array5by7 + 1
        co.columns['writtenaset']['3'] = array5by7 + 2
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 3
        co.commit('this is a commit message')
        co.close()

        co = aset_samples_initialized_repo.checkout(write=True)
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 3
        del co.columns['writtenaset']['1']
        del co.columns['writtenaset']['3']
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 1
        co.commit('this is a commit message')
        co.close()

        co = aset_samples_initialized_repo.checkout()
        with pytest.raises(KeyError):
            co.columns['writtenaset']['1']
        with pytest.raises(KeyError):
            co.columns['writtenaset']['3']
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 1
        assert_equal(co.columns['writtenaset']['2'], array5by7 + 1)
        co.close()

    def test_remove_data_multiple_items(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        co.columns['writtenaset']['1'] = array5by7
        co.columns['writtenaset']['2'] = array5by7 + 1
        co.columns['writtenaset']['3'] = array5by7 + 2
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 3
        co.commit('this is a commit message')
        co.close()

        co = aset_samples_initialized_repo.checkout(write=True)
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 3
        with pytest.raises(KeyError):
            del co.columns['writtenaset'][('1', '3')]
        assert '1' in co.columns['writtenaset']
        assert '3' in co.columns['writtenaset']
        del co.columns['writtenaset']['1']
        del co.columns['writtenaset']['3']
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 1
        co.commit('this is a commit message')
        co.close()

        co = aset_samples_initialized_repo.checkout()
        with pytest.raises(KeyError):
            co.columns['writtenaset']['1']
        with pytest.raises(KeyError):
            co.columns['writtenaset']['3']
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 1
        assert_equal(co.columns['writtenaset']['2'], array5by7 + 1)
        co.close()

    def test_pop_data(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        co.columns['writtenaset']['1'] = array5by7
        co.columns['writtenaset']['2'] = array5by7 + 1
        co.columns['writtenaset']['3'] = array5by7 + 2
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 3
        co.commit('this is a commit message')
        co.close()

        co = aset_samples_initialized_repo.checkout(write=True)
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 3
        res = co.columns['writtenaset'].pop('1')
        assert_equal(res, array5by7)

        aset = co.columns['writtenaset']
        res = aset.pop('3')
        assert_equal(res, array5by7 + 2)

        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 1
        co.commit('this is a commit message')
        co.close()

        co = aset_samples_initialized_repo.checkout()
        with pytest.raises(KeyError):
            co.columns['writtenaset']['1']
        with pytest.raises(KeyError):
            co.columns['writtenaset']['3']
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 1
        assert_equal(co.columns['writtenaset']['2'], array5by7 + 1)
        co.close()

    def test_pop_data_multiple_items(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        co.columns['writtenaset']['1'] = array5by7
        co.columns['writtenaset']['2'] = array5by7 + 1
        co.columns['writtenaset']['3'] = array5by7 + 2
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 3
        co.commit('this is a commit message')
        co.close()

        co = aset_samples_initialized_repo.checkout(write=True)
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 3
        with pytest.raises(TypeError):
            co.columns['writtenaset'].pop('1', '3')
        res = co.columns['writtenaset'].pop('1')
        assert_equal(res, array5by7)
        res = co.columns['writtenaset'].pop('3')
        assert_equal(res, array5by7 + 2)
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 1
        co.commit('this is a commit message')
        co.close()

        co = aset_samples_initialized_repo.checkout()
        with pytest.raises(KeyError):
            co.columns['writtenaset']['1']
        with pytest.raises(KeyError):
            co.columns['writtenaset']['3']
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 1
        assert_equal(co.columns['writtenaset']['2'], array5by7 + 1)
        co.close()

    def test_remove_all_data(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        co.columns['writtenaset']['1'] = array5by7
        new_array = np.zeros_like(array5by7)
        co.columns['writtenaset']['2'] = new_array
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 2
        co.commit('this is a commit message')
        co.close()

        co = aset_samples_initialized_repo.checkout(write=True)
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 2
        del co.columns['writtenaset']['1']
        del co.columns['writtenaset']['2']
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 0

        wset = co.columns['writtenaset']
        del co.columns['writtenaset']

        assert len(co.columns) == 0
        with pytest.raises(KeyError):
            len(co.columns['writtenaset'])
        co.commit('this is a commit message')
        co.close()

        # recreating same and verifying
        co = aset_samples_initialized_repo.checkout(write=True)
        assert len(co.columns) == 0
        co.add_ndarray_column('writtenaset', prototype=array5by7)
        co.columns['writtenaset']['1'] = array5by7
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 1
        co.commit('this is a commit message')
        co.close()

        co = aset_samples_initialized_repo.checkout()
        assert_equal(co.columns['writtenaset']['1'], array5by7)
        assert len(co.columns) == 1
        assert len(co.columns['writtenaset']) == 1
        co.close()

    def test_remove_data_nonexistant_sample_key_raises(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        co.columns['writtenaset']['1'] = array5by7
        new_array = np.zeros_like(array5by7)
        co.columns['writtenaset']['2'] = new_array
        co.columns['writtenaset']['3'] = new_array + 5
        with pytest.raises(KeyError):
            del co.columns['writtenaset']['doesnotexist']
        co.commit('this is a commit message')
        co.close()

    @pytest.mark.parametrize("aset1_backend", fixed_shape_backend_params)
    @pytest.mark.parametrize("aset2_backend", fixed_shape_backend_params)
    def test_multiple_columns_single_commit(
            self, aset1_backend, aset2_backend, aset_samples_initialized_repo, randomsizedarray
    ):
        co = aset_samples_initialized_repo.checkout(write=True)
        aset1 = co.add_ndarray_column('aset1', prototype=randomsizedarray, backend=aset1_backend)
        aset2 = co.add_ndarray_column('aset2', prototype=randomsizedarray, backend=aset2_backend)
        aset1['arr'] = randomsizedarray
        aset2['arr'] = randomsizedarray
        co.commit('this is a commit message')
        co.close()
        co = aset_samples_initialized_repo.checkout()
        assert_equal(co.columns['aset1']['arr'], randomsizedarray)
        assert_equal(co.columns['aset2']['arr'], randomsizedarray)
        co.close()

    @pytest.mark.parametrize("aset1_backend", fixed_shape_backend_params)
    @pytest.mark.parametrize("aset2_backend", fixed_shape_backend_params)
    def test_prototype_and_shape(self, aset1_backend, aset2_backend, repo, randomsizedarray):
        co = repo.checkout(write=True)
        aset1 = co.add_ndarray_column('aset1', prototype=randomsizedarray, backend=aset1_backend)
        aset2 = co.add_ndarray_column('aset2', shape=randomsizedarray.shape, dtype=randomsizedarray.dtype, backend=aset2_backend)

        newarray = np.random.random(randomsizedarray.shape).astype(randomsizedarray.dtype)
        aset1['arr1'] = newarray
        aset2['arr'] = newarray
        co.commit('this is a commit message')
        co.close()

        co = repo.checkout()
        assert_equal(co.columns['aset1']['arr1'], newarray)
        assert_equal(co.columns['aset2']['arr'], newarray)
        co.close()

    def test_samples_without_name(self, repo, randomsizedarray):
        co = repo.checkout(write=True)
        aset = co.add_ndarray_column('aset', prototype=randomsizedarray)
        with pytest.raises(TypeError):
            aset[randomsizedarray]

        aset_no_name = co.add_ndarray_column('aset_no_name', prototype=randomsizedarray)
        added = aset_no_name.append(randomsizedarray)
        assert_equal(next(aset_no_name.values()), randomsizedarray)
        assert_equal(aset_no_name[added], randomsizedarray)
        co.close()

    def test_append_samples(self, repo, randomsizedarray):
        co = repo.checkout(write=True)
        aset = co.add_ndarray_column('aset', prototype=randomsizedarray)
        with pytest.raises((ValueError, TypeError)):
            aset[randomsizedarray]

        aset_no_name = co.add_ndarray_column('aset_no_name', prototype=randomsizedarray)
        generated_key = aset_no_name.append(randomsizedarray)
        assert generated_key in aset_no_name
        assert len(aset_no_name) == 1
        assert_equal(aset_no_name[generated_key], randomsizedarray)
        co.close()

    def test_different_data_types_and_shapes(self, repo):
        co = repo.checkout(write=True)
        shape = (2, 3)
        dtype = np.int
        another_dtype = np.float64
        another_shape = (3, 4)
        arr = np.random.random(shape).astype(dtype)
        aset = co.add_ndarray_column('aset', shape=shape, dtype=dtype)
        aset['1'] = arr

        newarr = np.random.random(shape).astype(another_dtype)
        with pytest.raises(ValueError):
            aset['2'] = newarr

        newarr = np.random.random(another_shape).astype(dtype)
        with pytest.raises(ValueError):
            aset['3'] = newarr
        co.close()

    def test_add_sample_with_non_numpy_array_data_fails(self, aset_samples_initialized_repo):
        co = aset_samples_initialized_repo.checkout(write=True)
        with pytest.raises(ValueError, match='`data` argument type'):
            co.columns['writtenaset'][1] = [[1, 2, 3, 4, 5, 6, 7] for i in range(5)]
        co.close()

    def test_add_sample_with_fortran_order_data_fails(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        with pytest.raises(ValueError, match='`data` must be "C" contiguous array.'):
            co.columns['writtenaset'][1] = np.asfortranarray(array5by7)
        co.close()

    def test_add_sample_with_dimension_rank_fails(self, repo):
        co = repo.checkout(write=True)
        aset = co.add_ndarray_column('aset', shape=(2, 3), dtype=np.float32, variable_shape=True)
        arr = np.random.randn(2, 3, 2).astype(np.float32)
        with pytest.raises(ValueError, match='data rank 3 != aset rank 2'):
            aset[1] = arr
        co.close()

    def test_add_sample_with_dimension_exceeding_max_fails(self, repo):
        co = repo.checkout(write=True)
        aset = co.add_ndarray_column('aset', shape=(2, 3), dtype=np.float32, variable_shape=True)
        arr = np.random.randn(2, 4).astype(np.float32)
        with pytest.raises(ValueError, match='exceeds schema max'):
            aset[1] = arr
        co.close()

    @pytest.mark.parametrize("aset_backend", fixed_shape_backend_params)
    def test_writer_context_manager_column_add_sample(self, aset_backend, repo, randomsizedarray):
        co = repo.checkout(write=True)
        aset = co.add_ndarray_column('aset', prototype=randomsizedarray, backend=aset_backend)
        with co.columns['aset'] as aset:
            aset['1'] = randomsizedarray
        co.commit('this is a commit message')
        co.close()
        co = repo.checkout()
        assert_equal(co.columns['aset']['1'], randomsizedarray)
        co.close()

    @pytest.mark.parametrize("aset_backend", fixed_shape_backend_params)
    def test_column_context_manager_aset_sample_add(self, aset_backend, repo, randomsizedarray):
        co = repo.checkout(write=True)
        aset = co.add_ndarray_column('aset', prototype=randomsizedarray, backend=aset_backend)
        with co.columns['aset'] as aset:
            aset['1'] = randomsizedarray
            aset['2'] = randomsizedarray + 1
        co.commit('this is a commit message')
        co.close()

        co = repo.checkout()
        assert_equal(co.columns['aset']['1'], randomsizedarray)
        assert np.allclose(co.columns['aset'].get('2'), randomsizedarray + 1)
        co.close()

    def test_writer_column_properties_are_correct(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        assert co.columns.iswriteable is True
        d = co.columns['writtenaset']
        assert d.column =='writtenaset'
        assert d.dtype == array5by7.dtype
        assert np.allclose(d.shape, array5by7.shape) is True
        assert d.schema_type == 'fixed_shape'
        assert d.iswriteable is True
        assert d.backend == '01'
        assert isinstance(d.backend_options, dict)
        assert len(d.backend_options) > 0
        assert d.contains_subsamples is False
        assert d.remote_reference_keys == ()
        assert d.contains_remote_references is False
        co.close()

    def test_reader_column_properties_are_correct(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=False)
        assert co.columns.iswriteable is False
        d = co.columns['writtenaset']
        assert d.column =='writtenaset'
        assert d.dtype == array5by7.dtype
        assert np.allclose(d.shape, array5by7.shape) is True
        assert d.schema_type == 'fixed_shape'
        assert d.iswriteable is False
        assert d.backend == '01'
        assert isinstance(d.backend_options, dict)
        assert len(d.backend_options) > 0
        assert d.contains_subsamples is False
        assert d.remote_reference_keys == ()
        assert d.contains_remote_references is False

    def test_iter_column_samples_yields_keys(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        co.columns['writtenaset'][0] = array5by7
        new_array = np.zeros_like(array5by7)
        co.columns['writtenaset'][1] = new_array
        co.columns['writtenaset'][2] = new_array + 5

        for idx, sname in enumerate(iter(co.columns['writtenaset'])):
            assert sname == idx
        assert idx == 2
        co.close()

    def test_iter_columns_yields_aset_names(self, repo_20_filled_samples):
        co = repo_20_filled_samples.checkout(write=True)
        for k in iter(co.columns):
            assert k in ['second_aset', 'writtenaset']
        co.close()

    def test_set_item_column_fails(self, aset_samples_initialized_repo):
        co = aset_samples_initialized_repo.checkout(write=True)
        with pytest.raises(AttributeError):
            co.columns['newaset'] = co.columns['writtenaset']
        co.close()


class TestVariableSizedColumn(object):

    @pytest.mark.parametrize(
        'test_shapes,max_shape',
        [[[(2, 5), (1, 10), (10, 1), (5, 2)], (10, 10)],
         [[(10,), (10,)], (10,)],
         [[(3, 3, 3), (27, 1, 1), (1, 27, 1), (1, 1, 27), (3, 9, 1), (9, 3, 1), (1, 3, 9), (1, 9, 3)], (27, 27, 27)]])
    @pytest.mark.parametrize("dtype1", [np.uint8, np.float32, np.int32])
    @pytest.mark.parametrize("dtype2", [np.uint8, np.float32, np.int32])
    @pytest.mark.parametrize('backend1', variable_shape_backend_params)
    @pytest.mark.parametrize('backend2', variable_shape_backend_params)
    def test_write_all_zeros_same_size_different_shape_does_not_store_as_identical_hashs(
        self, repo, test_shapes, max_shape, dtype1, dtype2, backend1, backend2
    ):
        wco = repo.checkout(write=True)
        aset1 = wco.add_ndarray_column('aset1', shape=max_shape, dtype=dtype1, variable_shape=True, backend=backend1)
        aset2 = wco.add_ndarray_column('aset2', shape=max_shape, dtype=dtype2, variable_shape=True, backend=backend2)

        arrdict1, arrdict2 = {}, {}
        for idx, shape in enumerate(test_shapes):
            arr1 = np.zeros(shape, dtype=dtype1)
            arr2 = np.zeros(shape, dtype=dtype2)
            arrdict1[idx] = arr1
            arrdict2[idx] = arr2
            aset1[idx] = arr1
            aset2[idx] = arr2

        for k, v in arrdict1.items():
            # make sure they are good before committed
            res = aset1[k]
            assert res.dtype == v.dtype
            assert res.shape == v.shape
            assert_equal(res, v)
        for k, v in arrdict2.items():
            # make sure they are good before committed
            res = aset2[k]
            assert res.dtype == v.dtype
            assert res.shape == v.shape
            assert_equal(res, v)

        wco.commit('first')

        for k, v in arrdict1.items():
            # make sure they are good before committed
            res = aset1[k]
            assert res.dtype == v.dtype
            assert res.shape == v.shape
            assert_equal(res, v)
        for k, v in arrdict2.items():
            # make sure they are good before committed
            res = aset2[k]
            assert res.dtype == v.dtype
            assert res.shape == v.shape
            assert_equal(res, v)

        wco.close()
        rco = repo.checkout()
        naset1 = rco.columns['aset1']
        naset2 = rco.columns['aset2']

        for k, v in arrdict1.items():
            # make sure they are good before committed
            res = naset1[k]
            assert res.dtype == v.dtype
            assert res.shape == v.shape
            assert_equal(res, v)
        for k, v in arrdict2.items():
            # make sure they are good before committed
            res = naset2[k]
            assert res.dtype == v.dtype
            assert res.shape == v.shape
            assert_equal(res, v)
        rco.close()

    @pytest.mark.parametrize(
        'test_shapes,shape',
        [[[(10, 10), (1, 10), (2, 2), (3, 5), (1, 1), (10, 1)], (10, 10)],
         [[(10,), (1,), (5,)], (10,)],
         [[(100, 100, 100), (100, 100, 1), (100, 1, 100), (1, 100, 100), (1, 1, 1), (34, 6, 3)], (100, 100, 100)]])
    @pytest.mark.parametrize("dtype", [np.uint8, np.float32])
    @pytest.mark.parametrize('backend', variable_shape_backend_params)
    def test_writer_can_create_variable_size_column(
        self, aset_samples_initialized_repo, dtype, test_shapes, shape, backend
    ):
        repo = aset_samples_initialized_repo
        wco = repo.checkout(write=True)
        wco.add_ndarray_column('varaset', shape=shape, dtype=dtype, variable_shape=True, backend=backend)
        d = wco.columns['varaset']

        arrdict = {}
        for idx, shape in enumerate(test_shapes):
            arr = (np.random.random_sample(shape) * 10).astype(dtype)
            arrdict[str(idx)] = arr
            d[str(idx)] = arr

        for k, v in arrdict.items():
            # make sure they are good before committed
            assert_equal(d[k], v)

        wco.commit('first')

        for k, v in arrdict.items():
            # make sure they can work after commit
            assert_equal(d[k], v)
        wco.close()

    @pytest.mark.parametrize('test_shapes,shape', [
        [[(10, 10), (1, 10), (2, 2), (3, 5), (1, 1), (10, 1)], (10, 10)],
        [[(10,), (1,), (5,)], (10,)],
        [[(100, 100, 100), (100, 100, 1), (100, 1, 100), (1, 100, 100), (1, 1, 1), (34, 6, 3)], (100, 100, 100)]
    ])
    @pytest.mark.parametrize("dtype", [np.uint8, np.float32])
    @pytest.mark.parametrize('backend', variable_shape_backend_params)
    def test_reader_recieves_expected_values_for_variable_size_column(
        self, aset_samples_initialized_repo, dtype, test_shapes, shape, backend
    ):
        repo = aset_samples_initialized_repo
        wco = repo.checkout(write=True)
        wco.add_ndarray_column('varaset', shape=shape, dtype=dtype, variable_shape=True, backend=backend)
        wd = wco.columns['varaset']

        arrdict = {}
        for idx, shape in enumerate(test_shapes):
            arr = (np.random.random_sample(shape) * 10).astype(dtype)
            arrdict[str(idx)] = arr
            wd[str(idx)] = arr

        for k, v in arrdict.items():
            # make sure they are good before committed
            assert_equal(wd[k], v)

        wco.commit('first')
        rco = repo.checkout()
        rd = rco.columns['varaset']

        for k, v in arrdict.items():
            # make sure they can work after commit
            assert_equal(wd[k], v)
            assert_equal(rd[k], v)
        wco.close()
        rco.close()

    @pytest.mark.parametrize('aset_specs', [
        [['aset1', [(10, 10), (1, 10), (2, 2), (3, 5), (1, 1), (10, 1)], (10, 10)],
         ['aset2', [(10,), (1,), (5,)], (10,)]],
        [['aset1', [(100, 100), (1, 100), (20, 20), (30, 50), (1, 10), (10, 1)], (100, 100)],
         ['aset2', [(100,), (1,), (50,)], (100,)]]])
    @pytest.mark.parametrize('backends', permutations(variable_shape_backend_params, 2))
    @pytest.mark.parametrize('dtype', [np.float32, np.uint8])
    def test_writer_reader_can_create_read_multiple_variable_size_column(
        self, aset_samples_initialized_repo, aset_specs, backends, dtype
    ):
        repo = aset_samples_initialized_repo
        wco = repo.checkout(write=True)
        arrdict = {}
        for backend, aset_spec in zip(backends, aset_specs):
            aset_name, test_shapes, max_shape = aset_spec
            wco.add_ndarray_column(
                aset_name, shape=max_shape, dtype=dtype, variable_shape=True, backend=backend)

            arrdict[aset_name] = {}
            for idx, shape in enumerate(test_shapes):
                arr = (np.random.random_sample(shape) * 10).astype(dtype)
                arrdict[aset_name][str(idx)] = arr
                wco.columns[aset_name][str(idx)] = arr

        for aset_k in arrdict.keys():
            for samp_k, v in arrdict[aset_k].items():
                # make sure they are good before committed
                assert_equal(wco.columns[aset_k][samp_k], v)

        wco.commit('first')
        rco = repo.checkout()

        for aset_k in arrdict.keys():
            for samp_k, v in arrdict[aset_k].items():
                # make sure they are good before committed
                assert_equal(wco.columns[aset_k][samp_k], v)
                assert_equal(rco.columns[aset_k][samp_k], v)
        wco.close()
        rco.close()

    def test_writer_column_properties_are_correct(self, aset_samples_var_shape_initialized_repo):
        co = aset_samples_var_shape_initialized_repo.checkout(write=True)
        d = co.columns['writtenaset']
        assert d.column =='writtenaset'
        assert d.dtype == np.float64
        assert np.allclose(d.shape, (10, 10))
        assert d.schema_type == 'variable_shape'
        assert d.iswriteable is True
        assert d.backend in variable_shape_backend_params
        assert isinstance(d.backend_options, dict)
        assert d.contains_subsamples is False
        assert d.remote_reference_keys == ()
        assert d.contains_remote_references is False
        co.close()

    def test_reader_column_properties_are_correct(self, aset_samples_var_shape_initialized_repo):
        co = aset_samples_var_shape_initialized_repo.checkout(write=False)
        d = co.columns['writtenaset']
        assert d.column =='writtenaset'
        assert d.dtype == np.float64
        assert np.allclose(d.shape, (10, 10))
        assert d.schema_type == 'variable_shape'
        assert d.iswriteable is False
        assert d.backend in variable_shape_backend_params
        assert isinstance(d.backend_options, dict)
        assert d.contains_subsamples is False
        assert d.remote_reference_keys == ()
        assert d.contains_remote_references is False
        co.close()


class TestMultiprocessColumnReads(object):

    @pytest.mark.parametrize('backend', fixed_shape_backend_params)
    def test_external_multi_process_pool(self, repo, backend):
        from multiprocessing import get_context

        masterCmtList = []
        co = repo.checkout(write=True)
        co.add_ndarray_column(name='writtenaset', shape=(20, 20), dtype=np.float32, backend=backend)
        masterSampList = []
        for cIdx in range(2):
            if cIdx != 0:
                co = repo.checkout(write=True)
            with co.columns['writtenaset'] as d:
                kstart = 20 * cIdx
                for sIdx in range(20):
                    arr = np.random.randn(20, 20).astype(np.float32) * 100
                    sName = str(sIdx + kstart)
                    d[sName] = arr
                    masterSampList.append(arr)
            assert d.backend == backend
            cmt = co.commit(f'master commit number: {cIdx}')
            masterCmtList.append((cmt, list(masterSampList)))
            co.close()

        cmtIdx = 0
        for cmt, sampList in masterCmtList:
            nco = repo.checkout(write=False, commit=cmt)
            ds = nco.columns['writtenaset']
            keys = [str(i) for i in range(20 + (20*cmtIdx))]
            with get_context().Pool(2) as P:
                cmtData = P.map(ds.get, keys)
            for data, sampData in zip(cmtData, sampList):
                assert_equal(data, sampData) is True
            cmtIdx += 1
            nco.close()

    @pytest.mark.parametrize('backend', fixed_shape_backend_params)
    def test_external_multi_process_pool_fails_on_write_enabled_checkout(self, repo, backend):
        from multiprocessing import get_context

        co = repo.checkout(write=True)
        co.add_ndarray_column(name='writtenaset', shape=(20, 20), dtype=np.float32, backend=backend)
        with co.columns['writtenaset'] as d:
            for sIdx in range(20):
                d[sIdx] = np.random.randn(20, 20).astype(np.float32) * 100
        assert d.backend == backend
        co.commit(f'master commit number 1')
        co.close()

        nco = repo.checkout(write=True)
        ds = nco.columns['writtenaset']
        keys = [i for i in range(20)]
        with pytest.raises(PermissionError):
            with get_context().Pool(2) as P:
                cmtData = P.map(ds.get, keys)
        nco.close()


    @pytest.mark.parametrize('backend', fixed_shape_backend_params)
    def test_multiprocess_get_succeeds_on_superset_and_subset_of_keys(self, repo, backend):
        from multiprocessing import get_context

        co = repo.checkout(write=True)
        co.add_ndarray_column(name='writtenaset', shape=(20, 20), dtype=np.float32, backend=backend)
        masterSampList = []
        with co.columns['writtenaset'] as d:
            for sIdx in range(20):
                arr = np.random.randn(20, 20).astype(np.float32) * 100
                d[sIdx] = arr
                masterSampList.append(arr)
            assert d.backend == backend
        cmt = co.commit(f'master commit number one')
        co.close()

        nco = repo.checkout(write=False, commit=cmt)
        ds = nco.columns['writtenaset']

        # superset of keys fails
        keys = [i for i in range(24)]
        with get_context().Pool(2) as P:
            cmtData = P.map(ds.get, keys)
        for idx, data in enumerate(cmtData):
            if idx >= 20:
                assert data is None
            else:
                assert_equal(data, masterSampList[idx])

        # subset of keys works
        keys = [i for i in range(10, 20)]
        with get_context().Pool(2) as P:
            cmtData = P.map(ds.get, keys)
        for idx, data in enumerate(cmtData):
            assert_equal(data, masterSampList[10+idx])
        nco.close()

    def test_writer_iterating_over_keys_can_have_additions_made_no_error(self, two_commit_filled_samples_repo):
        # do not want ``RuntimeError dictionary changed size during iteration``

        repo = two_commit_filled_samples_repo
        co = repo.checkout(write=True)
        aset = co.columns['writtenaset']
        with aset as ds:
            for idx, k in enumerate(ds.keys()):
                if idx == 0:
                    ds['1232'] = np.random.randn(5, 7).astype(np.float32)
                assert '1232' != k

        added_key_exists_on_later_iteration = False
        for k in aset.keys():
            if k == '1232':
                added_key_exists_on_later_iteration = True
                break
        assert added_key_exists_on_later_iteration is True
        co.close()

    def test_writer_iterating_over_values_can_have_additions_made_no_error(self, two_commit_filled_samples_repo):
        # do not want ``RuntimeError dictionary changed size during iteration``

        repo = two_commit_filled_samples_repo
        co = repo.checkout(write=True)
        aset = co.columns['writtenaset']
        mysample = np.random.randn(5, 7).astype(np.float32)
        with aset as ds:
            for idx, v in enumerate(ds.values()):
                if idx == 0:
                    ds['1232'] = mysample
                assert not np.allclose(v, mysample)

        added_value_exists_on_later_iteration = False
        for v in aset.values():
            if np.allclose(v, mysample):
                added_value_exists_on_later_iteration = True
                break
        assert added_value_exists_on_later_iteration is True
        co.close()

    def test_writer_iterating_over_items_can_have_additions_made_no_error(self, two_commit_filled_samples_repo):
        # do not want ``RuntimeError dictionary changed size during iteration``

        repo = two_commit_filled_samples_repo
        co = repo.checkout(write=True)
        aset = co.columns['writtenaset']
        mysample = np.random.randn(5, 7).astype(np.float32)
        with aset as ds:
            for idx, kv in enumerate(ds.items()):
                if idx == 0:
                    ds['1232'] = mysample
                k, v = kv
                assert not np.allclose(v, mysample)
                assert k != '1232'

        added_value_exists_on_later_iteration = False
        for k, v in aset.items():
            if (k == '1232') and np.allclose(v, mysample):
                added_value_exists_on_later_iteration = True
                break
        assert added_value_exists_on_later_iteration is True
        co.close()

    def test_reader_iterating_over_items_can_not_make_additions(self, two_commit_filled_samples_repo):
        # do not want ``RuntimeError dictionary changed size during iteration``

        repo = two_commit_filled_samples_repo
        co = repo.checkout(write=False)
        aset = co.columns['writtenaset']
        mysample = np.random.randn(5, 7).astype(np.float32)
        with aset as ds:
            for idx, kv in enumerate(ds.items()):
                if idx == 0:
                    with pytest.raises(TypeError):
                        ds['1232'] = mysample
                k, v = kv
                assert not np.allclose(v, mysample)
                assert k != '1232'

        assert '1232' not in aset
        co.close()
